from fairseq.tasks.speech_to_text import SpeechToTextTask
from fairseq.data import FairseqDataset, data_utils, ResamplingDataset, ConcatDataset
from .Fairseq_SpeechToTextDataset import SpeechToTextDataset, SpeechToTextDatasetCreator, S2TDataConfig
from typing import Dict, List, Optional, Tuple
import torchaudio
import torch
from argparse import Namespace
import os.path as op
import csv


# the following codes are modify from fairseq's implementation
# (https://github.com/pytorch/fairseq)
class S3prl_SpeechToTextTask(SpeechToTextTask):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load_dataset(self, split, max_feature_len = -1, epoch=1, combine=False, **kwargs):
        is_train_split = split.startswith("train")
        pre_tokenizer = self.build_tokenizer(self.args)
        bpe_tokenizer = self.build_bpe(self.args)
        self.datasets[split] = S3prl_SpeechToTextDatasetCreator.from_tsv(
            self.args.data,
            self.data_cfg,
            split,
            self.tgt_dict,
            pre_tokenizer,
            bpe_tokenizer,
            is_train_split=is_train_split,
            epoch=epoch,
            seed=self.args.seed,
            upstream_rate = self.upstream_rate,
            max_feature_len = max_feature_len,
        )
    
    def build_model(self, args, input_dim):
        args.input_feat_per_channel = input_dim
        args.input_channels = self.data_cfg.input_channels
        return super(SpeechToTextTask, self).build_model(args)

class S3prl_SpeechToTextDatasetCreator(SpeechToTextDatasetCreator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # optional
    KEY_SAMPLE_RATE = 'sr'

    # default
    DEFAULT_SAMPLE_RATE = 16000

    @classmethod
    def _from_list(
        cls,
        split_name: str,
        is_train_split,
        samples: List[List[Dict]],
        data_cfg: S2TDataConfig,
        tgt_dict,
        pre_tokenizer,
        bpe_tokenizer,
        upstream_rate,
        max_feature_len,
    ) -> SpeechToTextDataset:
        audio_paths, n_frames, src_texts, tgt_texts, ids = [], [], [], [], []
        speakers, src_langs, tgt_langs = [], [], []
        srs = []
        for s in samples:
            ids.extend([ss[cls.KEY_ID] for ss in s])
            audio_paths.extend(
                [op.join(data_cfg.audio_root, ss[cls.KEY_AUDIO]) for ss in s]
            )
            n_frames.extend([int(ss[cls.KEY_N_FRAMES]) for ss in s])
            tgt_texts.extend([ss[cls.KEY_TGT_TEXT] for ss in s])
            src_texts.extend(
                [ss.get(cls.KEY_SRC_TEXT, cls.DEFAULT_SRC_TEXT) for ss in s]
            )
            speakers.extend([ss.get(cls.KEY_SPEAKER, cls.DEFAULT_SPEAKER) for ss in s])
            src_langs.extend([ss.get(cls.KEY_SRC_LANG, cls.DEFAULT_LANG) for ss in s])
            tgt_langs.extend([ss.get(cls.KEY_TGT_LANG, cls.DEFAULT_LANG) for ss in s])
            
            # sample rate
            srs.extend([int(ss.get(cls.KEY_SAMPLE_RATE, cls.DEFAULT_SAMPLE_RATE)) for ss in s])
        
        return S3prl_SpeechToTextDataset(
            split_name,
            is_train_split,
            data_cfg,
            audio_paths,
            n_frames,
            src_texts = src_texts,
            tgt_texts = tgt_texts,
            speakers = speakers,
            src_langs = src_langs,
            tgt_langs = tgt_langs,
            srs = srs,
            ids = ids,
            tgt_dict = tgt_dict,
            pre_tokenizer = pre_tokenizer,
            bpe_tokenizer = bpe_tokenizer,
            upstream_rate = upstream_rate,
            max_feature_len = max_feature_len,
        )

    @classmethod
    def from_tsv(
        cls,
        root: str,
        data_cfg: S2TDataConfig,
        splits: str,
        tgt_dict,
        pre_tokenizer,
        bpe_tokenizer,
        is_train_split: bool,
        epoch: int,
        seed: int,
        upstream_rate: int,
        max_feature_len: int,
    ) -> SpeechToTextDataset:

        _splits = splits.split(",")
        assert len(_splits) == 1, "do not support multiple files training"

        split = _splits[0]
        tsv_path = op.join(root, f"{split}.tsv")
        if not op.isfile(tsv_path):
            raise FileNotFoundError(f"Dataset not found: {tsv_path}")
        
        with open(tsv_path) as f:
            reader = csv.DictReader(
                f,
                delimiter="\t",
                quotechar=None,
                doublequote=False,
                lineterminator="\n",
                quoting=csv.QUOTE_NONE,
            )
            samples = [dict(e) for e in reader]

        assert len(samples) > 0

        return cls._from_list(
            split,
            is_train_split,
            [samples],
            data_cfg,
            tgt_dict,
            pre_tokenizer,
            bpe_tokenizer,
            upstream_rate,
            max_feature_len,
        )


class S3prl_SpeechToTextDataset(SpeechToTextDataset):

    TARGET_RATE = 16000

    def __init__(self, *args, srs = Optional[List[int]], upstream_rate = 160, max_feature_len=-1, **kwargs):
        
        super().__init__(*args, **kwargs)

        self.srs = srs
        self.max_feature_len = max_feature_len
        self.max_wav_len = max_feature_len * upstream_rate
        self.resamplers = {}
        for sr in set(srs):
            self.resamplers[sr] = torchaudio.transforms.Resample(
                orig_freq = sr,
                new_freq = self.TARGET_RATE,
            )

        for i in range(len(self.n_frames)):
            new_n_frames = self.n_frames[i] * self.TARGET_RATE / self.srs[i] / upstream_rate
            if self.max_feature_len > 0 and new_n_frames > max_feature_len:
                new_n_frames = max_feature_len
            self.n_frames[i] = int(new_n_frames)

    def __getitem__(
        self, index: int
    ) -> Tuple[str, int, torch.Tensor, Optional[torch.Tensor]]:

        source, sr = torchaudio.load(self.audio_paths[index])

        assert self.srs[index] == sr
        source = self.resamplers[sr](source)
        source = torch.mean(source, dim=0)
        source = source.view(-1)

        if self.feature_transforms is not None:
            assert not self.data_cfg.use_audio_input
            source = self.feature_transforms(source)
        source = source.float()

        # truncate the wav
        if self.max_feature_len > 0:
            if source.size(0) > self.max_wav_len:
                print(f'wav too long({source.size(0)}), truncate to {self.max_wav_len} (id={index})')
                source = source[:self.max_wav_len]

        target = None
        if self.tgt_texts is not None:
            tokenized = self.tokenize_text(self.tgt_texts[index])
            target = self.tgt_dict.encode_line(
                tokenized, add_if_not_exist=False, append_eos=True
            ).long()
            if self.data_cfg.prepend_tgt_lang_tag:
                lang_tag = self.LANG_TAG_TEMPLATE.format(self.tgt_langs[index])
                lang_tag_idx = self.tgt_dict.index(lang_tag)
                target = torch.cat((torch.LongTensor([lang_tag_idx]), target), 0)
        return self.ids[index], index, source, target

    def collater(self, samples: List[Tuple[str, int, torch.Tensor, torch.Tensor]]):
        ids = [sample[0] for sample in samples]
        samples = [sample[1:] for sample in samples]
        output_dict = super().collater(samples)
        output_dict['utt_id'] = ids

        wavs = []

        for i in range(output_dict['nsentences']):
            wav = output_dict['net_input']['src_tokens'][i]
            length = output_dict['net_input']['src_lengths'][i].item()
            wavs.append(wav[:length].numpy())

        return wavs, output_dict



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ utility/run_sig_test.py ]
#   Synopsis     [ Run the significance test on TWO RELATED samples of scores, a and b, from two experiment checkpoints ]
#   Author       [ Andy T. Liu (https://github.com/andi611) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


"""
Note:
In order to calculate the significance tests,
the downstream expert's forward() method should log a metric score for each testing sample.
The `records['sample_wise_metric']` should be a list containing the testing result of each sample.

For example:
```
python
# for frame-wise classification
for sample in samples:
    records['sample_wise_metric'] += [torch.FloatTensor(sample).mean().item()]
# for utterance-wise classification
records['sample_wise_metric'] += (predicted_classid == labels).view(-1).cpu().tolist()
```
"""


###############
# IMPORTATION #
###############
import os
import glob
import random
import argparse
from tqdm import tqdm
from argparse import Namespace
#-------------#
import torch
import torchaudio
import numpy as np
from statsmodels.stats.contingency_tables import mcnemar
from scipy import stats
#-------------#
from s3prl.downstream.runner import Runner
from s3prl.utility.helper import hack_isinstance, override, defaultdict


def get_ttest_args():
    parser = argparse.ArgumentParser()
    # ttest, for continuous result: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ttest_rel.html#scipy.stats.ttest_rel
    # fisher, for categorical results: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.fisher_exact.html
    # mcnemar, for categorical results: https://www.statsmodels.org/dev/generated/statsmodels.stats.contingency_tables.mcnemar.html
    parser.add_argument('-m', '--mode', choices=['ttest', 'fisher', 'mcnemar'], default='ttest')
    
    parser.add_argument('-em', '--evaluate_metric', default='acc')
    parser.add_argument('-t', '--evaluate_split', default='test')
    parser.add_argument('-o', '--override', help='Used to override args and config, this is at the highest priority')

    # compare two ckpts with the Paired Sample T-test using SciPy
    # All the args and config ill be determined by the ckpts
    # if a directory is specified, the latest ckpt will be used by default
    parser.add_argument('-e1', '--past_exp1', metavar='{CKPT_PATH,CKPT_DIR}', help='Load from a checkpoint')
    parser.add_argument('-e2', '--past_exp2', metavar='{CKPT_PATH,CKPT_DIR}', help='Load from another checkpoint')
    parser.add_argument('-u1', '--upstream1', default='default', type=str, help='used to override the upstream string for checkpoint e1')
    parser.add_argument('-u2', '--upstream2', default='default', type=str, help='used to override the upstream string for checkpoint e2')

    # options
    parser.add_argument('--seed', default=1337, type=int)
    parser.add_argument('--verbose', action='store_true', help='Print model infomation')
    parser.add_argument('--ckpt_name', default='best-states-dev', \
                        help='The string used for searching the checkpoint, \
                        example choices: `states-*`, `best-states-dev`, `best-states-test`.')
    args = parser.parse_args()

    args1, config1 = get_past_exp(args, args.past_exp1, args.ckpt_name)
    args2, config2 = get_past_exp(args, args.past_exp2, args.ckpt_name)
    if args.upstream1 != 'default': args1.upstream = args.upstream1
    if args.upstream2 != 'default': args2.upstream = args.upstream2

    return args.mode, args1, config1, args2, config2


def get_past_exp(args, past_exp, name):
    # determine checkpoint path
    if os.path.isdir(past_exp):
        ckpt_pths = glob.glob(os.path.join(past_exp, f'{name}.ckpt'))
        assert len(ckpt_pths) > 0
        if len(ckpt_pths) == 1:
            ckpt_pth = ckpt_pths[0]
        else:
            ckpt_pths = sorted(ckpt_pths, key=lambda pth: int(pth.split('-')[-1].split('.')[0]))
            ckpt_pth = ckpt_pths[-1]
    else:
        ckpt_pth = past_exp

    print(f'[Runner] - Loading from {ckpt_pth}')

    # load checkpoint
    ckpt = torch.load(ckpt_pth, map_location='cpu')

    def update_args(old, new, preserve_list=None):
        out_dict = vars(old)
        new_dict = vars(new)
        for key in list(new_dict.keys()):
            if key in preserve_list:
                new_dict.pop(key)
        out_dict.update(new_dict)
        return Namespace(**out_dict)

    # overwrite args
    cannot_overwrite_args = [
        'mode', 'evaluate_split', 'override',
        'backend', 'local_rank', 'past_exp',
    ]
    args = update_args(args, ckpt['Args'], preserve_list=cannot_overwrite_args)

    args.init_ckpt = ckpt_pth
    args.mode = 'evaluate'
    config = ckpt['Config']

    if args.override:
        override(args.override, args, config)
    return args, config


class Tester(Runner):
    """
    Used to handle the evaluation loop and return the testing records for Paired Sample T-test.
    """
    def __init__(self, args, config):
        super(Tester, self).__init__(args, config)
    
    def evaluate(self):
        """evaluate function will always be called on a single process even during distributed training"""

        split = self.args.evaluate_split

        # fix seed to guarantee the same evaluation protocol across steps 
        random.seed(self.args.seed)
        np.random.seed(self.args.seed)
        torch.manual_seed(self.args.seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(self.args.seed)
        with torch.cuda.device(self.args.device):
            torch.cuda.empty_cache()

        # set all models to eval
        self.downstream.eval()
        self.upstream.eval()

        # prepare data
        dataloader = self.downstream.get_dataloader(split)

        records = defaultdict(list)
        for batch_id, (wavs, *others) in enumerate(tqdm(dataloader, dynamic_ncols=True, desc=split)):

            wavs = [torch.FloatTensor(wav).to(self.args.device) for wav in wavs]
            with torch.no_grad():
                features = self.upstream(wavs)
                self.downstream(
                    split,
                    features, *others,
                    records = records,
                )
        return records


def process_records(records, metric):
    assert 'sample_wise_metric' in records, 'Utterance-wise / sample-wise metric is necessary for proceeding the Paired Sample T-test.'
    average = torch.FloatTensor(records[metric]).mean().item()
    return average, records['sample_wise_metric']


def main():
    torch.multiprocessing.set_sharing_strategy('file_system')
    torchaudio.set_audio_backend('sox_io')
    hack_isinstance()

    # get config and arguments
    mode, args1, config1, args2, config2 = get_ttest_args()

    # Fix seed and make backends deterministic
    random.seed(args1.seed)
    np.random.seed(args1.seed)
    torch.manual_seed(args1.seed)
    if torch.cuda.is_available(): torch.cuda.manual_seed_all(args1.seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    tester1 = Tester(args1, config1)
    records1 = eval(f'tester1.{args1.mode}')()
    average1, sample_metric1 = process_records(records1, args1.evaluate_metric)

    tester2 = Tester(args2, config2)
    records2 = eval(f'tester2.{args2.mode}')()
    average2, sample_metric2 = process_records(records2, args2.evaluate_metric)

    if mode == 'ttest':
        statistic, p_value = stats.ttest_rel(sample_metric1, sample_metric2)
    elif mode == 'fisher':
        correct1 = sample_metric1.count(True)
        correct2 = sample_metric2.count(True)
        contingency_table = [[correct1, correct2], 
                             [len(sample_metric1)-correct1, len(sample_metric2)-correct2]]
        statistic, p_value = stats.fisher_exact(contingency_table)
    elif mode == 'mcnemar':
        correct1 = sample_metric1.count(True)
        correct2 = sample_metric2.count(True)
        contingency_table = [[correct1, correct2], 
                             [len(sample_metric1)-correct1, len(sample_metric2)-correct2]]
        b = mcnemar(contingency_table, exact=True)
        statistic, p_value = b.statistic, b.pvalue
    else:
        raise NotImplementedError
    
    print(f'[Runner] - The testing scores of the two ckpts are {average1} and {average2}, respectively.')
    print(f'[Runner] - The statistic of the significant test of the two ckpts is {statistic}')
    print(f'[Runner] - The P value of significant test of the two ckpts is {p_value}')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ pretrain/bucket_dataset.py ]
#   Synopsis     [ the general acoustic dataset with bucketing ]
#   Author1      [ Andy T. Liu (https://github.com/andi611) ]
#   Author2      [ Heng-Jui Chang (https://github.com/vectominist) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import os
import random
import pandas as pd
from torch.utils.data.dataset import Dataset


HALF_BATCHSIZE_TIME = 99999


################
# FEAT DATASET #
################
class FeatDataset(Dataset):
    """Base On-the-fly feature dataset by Andy T. Liu"""
    
    def __init__(self, extracter, task_config, bucket_size, file_path, sets, 
                 max_timestep=0, libri_root=None, **kwargs):
        super(FeatDataset, self).__init__()

        self.extracter = extracter
        self.task_config = task_config
        self.libri_root = libri_root
        self.sample_length = task_config['sequence_length'] 
        if self.sample_length > 0:
            print('[Dataset] - Sampling random segments for training, sample length:', self.sample_length)
        
        # Read file
        self.root = file_path
        tables = [pd.read_csv(os.path.join(file_path, s + '.csv')) for s in sets]
        self.table = pd.concat(tables, ignore_index=True).sort_values(by=['length'], ascending=False)
        print('[Dataset] - Training data from these sets:', str(sets))

        # Drop seqs that are too long
        if max_timestep > 0:
            self.table = self.table[self.table.length < max_timestep]
        # Drop seqs that are too short
        if max_timestep < 0:
            self.table = self.table[self.table.length > (-1 * max_timestep)]

        X = self.table['file_path'].tolist()
        X_lens = self.table['length'].tolist()
        self.num_samples = len(X)
        print('[Dataset] - Number of individual training instances:', self.num_samples)

        # Use bucketing to allow different batch size at run time
        self.X = []
        batch_x, batch_len = [], []

        for x, x_len in zip(X, X_lens):
            batch_x.append(x)
            batch_len.append(x_len)
            
            # Fill in batch_x until batch is full
            if len(batch_x) == bucket_size:
                # Half the batch size if seq too long
                if (bucket_size >= 2) and (max(batch_len) > HALF_BATCHSIZE_TIME) and self.sample_length == 0:
                    self.X.append(batch_x[:bucket_size//2])
                    self.X.append(batch_x[bucket_size//2:])
                else:
                    self.X.append(batch_x)
                batch_x, batch_len = [], []
        
        # Gather the last batch
        if len(batch_x) > 1: 
            self.X.append(batch_x)

    def _sample(self, x):
        if self.sample_length <= 0: return x
        if len(x) < self.sample_length: return x
        idx = random.randint(0, len(x)-self.sample_length)
        return x[idx:idx+self.sample_length]

    def __len__(self):
        return len(self.X)

    def collate_fn(self, items):
        items = items[0] # hack bucketing
        return items


################
# WAVE DATASET #
################
class WaveDataset(Dataset):
    """Base waveform dataset for Disiller by Heng-Jui Chang"""

    def __init__(
        self,
        task_config,
        bucket_size,
        file_path,
        sets,
        max_timestep=0,
        libri_root=None,
        **kwargs
    ):
        super().__init__()

        self.task_config = task_config
        self.libri_root = libri_root
        self.sample_length = task_config["sequence_length"]
        if self.sample_length > 0:
            print(
                "[Dataset] - Sampling random segments for training, sample length:",
                self.sample_length,
            )

        # Read file
        self.root = file_path
        tables = [pd.read_csv(os.path.join(file_path, s + ".csv")) for s in sets]
        self.table = pd.concat(tables, ignore_index=True).sort_values(
            by=["length"], ascending=False
        )
        print("[Dataset] - Training data from these sets:", str(sets))

        # Drop seqs that are too long
        if max_timestep > 0:
            self.table = self.table[self.table.length < max_timestep]
        # Drop seqs that are too short
        if max_timestep < 0:
            self.table = self.table[self.table.length > (-1 * max_timestep)]

        X = self.table["file_path"].tolist()
        X_lens = self.table["length"].tolist()
        self.num_samples = len(X)
        print("[Dataset] - Number of individual training instances:", self.num_samples)

        # Use bucketing to allow different batch size at run time
        self.X = []
        batch_x, batch_len = [], []

        for x, x_len in zip(X, X_lens):
            batch_x.append(x)
            batch_len.append(x_len)

            # Fill in batch_x until batch is full
            if len(batch_x) == bucket_size:
                # Half the batch size if seq too long
                if (
                    (bucket_size >= 2)
                    and (max(batch_len) > HALF_BATCHSIZE_TIME)
                    and self.sample_length == 0
                ):
                    self.X.append(batch_x[: bucket_size // 2])
                    self.X.append(batch_x[bucket_size // 2 :])
                else:
                    self.X.append(batch_x)
                batch_x, batch_len = [], []

        # Gather the last batch
        if len(batch_x) > 1:
            self.X.append(batch_x)

    def _sample(self, x):
        if self.sample_length <= 0:
            return x
        if len(x) < self.sample_length:
            return x
        idx = random.randint(0, len(x) - self.sample_length)
        return x[idx : idx + self.sample_length]

    def __len__(self):
        return len(self.X)

    def collate_fn(self, items):
        items = items[0]  # hack bucketing
        assert (
            len(items) == 4
        ), "__getitem__ should return (wave_input, wave_orig, wave_len, pad_mask)"
        return items


# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/mockingjay/expert.py ]
#   Synopsis     [ the mockingjay wrapper ]
#   Author       [ Andy T. Liu (https://github.com/andi611) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


from typing import List, Tuple
from collections import OrderedDict

import yaml
import torch
from torch import Tensor

from ..interfaces import UpstreamBase
from .builder import PretrainedTransformer


class UpstreamExpert(UpstreamBase):
    """
    The Mockingjay wrapper
    """

    def __init__(self, ckpt, options_config=None, **kwargs):
        super().__init__(**kwargs)

        if options_config is not None:
            print(
                "[UpstreamExpert] - Using upstream expert config file from:",
                options_config,
            )
            with open(options_config, "r") as file:
                options = yaml.load(file, Loader=yaml.FullLoader)
        else:
            print("[UpstreamExpert] - Using the default upstream expert config")
            options = {
                "load_pretrain": "True",
                "no_grad": "False",
                "dropout": "default",
                "spec_aug": "False",
                "spec_aug_prev": "True",
                "output_hidden_states": "True",
                "permute_input": "False",
            }

        options["ckpt_file"] = ckpt
        options["select_layer"] = -1

        self.transformer = PretrainedTransformer(options, inp_dim=-1)
        assert hasattr(
            self.transformer, "extracter"
        ), "This wrapper only supports `on-the-fly` ckpt with built in feature extracters."
        self.transformer([torch.randn(16000)])

    def get_downsample_rates(self, key: str) -> int:
        return 160

    def forward(self, wavs):
        last_hidden_state, hidden_states = self.transformer(wavs)  # (batch_size, extracted_seqlen, feature_dim)
        return {
            "last_hidden_state": last_hidden_state,
            "hidden_states": hidden_states.unbind(dim=0),
        }



import sys
import yaml
import argparse
import torch

parser = argparse.ArgumentParser()
parser.add_argument('--ckpt_pth', required=True)
parser.add_argument('--settings_dir', required=True)
parser.add_argument('--new_ckpt_pth', required=True)
paras = parser.parse_args()

ckpt = torch.load(paras.ckpt_pth, map_location='cpu')

with open(f'{paras.settings_dir}/args.yaml', 'r') as handle:
    args = yaml.load(handle, Loader=yaml.FullLoader)
    ckpt['Args'] = argparse.Namespace(**args)

with open(f'{paras.settings_dir}/config.yaml', 'r') as handle:
    config = yaml.load(handle, Loader=yaml.FullLoader)
    ckpt['Config'] = config

torch.save(ckpt, paras.new_ckpt_pth)



from pase.models.frontend import wf_builder
pase = wf_builder('cfg/frontend/PASE+.cfg').eval()
pase.load_pretrained('FE_e199.ckpt', load_last=True, verbose=True)
pase.cuda()

import sys
wav_path = sys.argv[1]
out_path = sys.argv[2]

# Now we can forward waveforms as Torch tensors
import torch
import torchaudio
torchaudio.set_audio_backend('sox')

x, sr = torchaudio.load(wav_path)
x = x.view(-1).cuda()

x = x.view(1, 1, -1)

with torch.no_grad():
    pase.eval()

    # y size will be (1, 256, 625), which are 625 frames of 256 dims each
    y = pase(x)[0].transpose(0, 1)

torch.save(y.detach().cpu(), out_path)



import random
import re
import xml.etree.ElementTree as ET
from collections import defaultdict
from pathlib import Path

import torch
from torch.utils.data.dataset import Dataset
from torchaudio.sox_effects import apply_effects_file, apply_effects_tensor
from tqdm import tqdm


class SWS2013Dataset(Dataset):
    """SWS 2013 dataset."""

    def __init__(self, split, **kwargs):
        assert split in ["dev", "eval"]

        dataset_root = Path(kwargs["sws2013_root"])
        split_root = Path(kwargs["sws2013_scoring_root"]) / f"sws2013_{split}"

        # parse infos
        audio2dur = parse_ecf(split_root / "sws2013.ecf.xml")
        query2audios = parse_rttm(split_root / f"sws2013_{split}.rttm")
        query2tensors = find_queries(dataset_root / f"{split}_queries")
        print(f"[SWS2013] # of audios: {len(audio2dur)}")
        print(f"[SWS2013] # of queries: {len(query2tensors)}")

        # find complement set
        all_audio_set = set(audio2dur.keys())
        query2audio_set = {
            query: set(audio_info["audio"] for audio_info in audio_infos)
            for query, audio_infos in query2audios.items()
        }
        query2audio_compl_set = {
            query: all_audio_set - audio_set
            for query, audio_set in query2audio_set.items()
        }

        # form positive & negative pairs
        positive_pairs, negative_pairs = [], []
        for query, tensors in query2tensors.items():
            for query_tensor in tensors:
                negative_pairs.append(
                    {
                        "query_tensor": query_tensor,
                        "audio_set": query2audio_compl_set[query]
                        if query in query2audio_compl_set
                        else all_audio_set,
                    }
                )
                if query not in query2audios:
                    continue
                for audio_info in query2audios[query]:
                    positive_pairs.append(
                        {
                            "query_tensor": query_tensor,
                            "audio": audio_info["audio"],
                            "offset": audio_info["offset"],
                            "duration": audio_info["duration"],
                        }
                    )

        print(f"[SWS2013] # of positive pairs: {len(positive_pairs)}")

        self.audio_dir = dataset_root / "Audio"
        self.audio2dur = audio2dur
        self.max_dur = 3.0
        self.positive_pairs = positive_pairs
        self.negative_pairs = negative_pairs

    def __len__(self):
        return len(self.positive_pairs) + len(self.negative_pairs)

    def __getitem__(self, idx):
        if idx < len(self.positive_pairs):  # positive pair
            pair = self.positive_pairs[idx]
            audio_path = (self.audio_dir / pair["audio"]).with_suffix(".wav")
            audio_tensor = path2segment(
                audio_path, pair["duration"], self.max_dur, pair["offset"]
            )
        else:  # negative pair
            pair = self.negative_pairs[idx - len(self.positive_pairs)]
            sample_audio = random.sample(pair["audio_set"], 1)[0]
            audio_dur = self.audio2dur[sample_audio]
            audio_path = (self.audio_dir / sample_audio).with_suffix(".wav")
            audio_tensor = path2segment(audio_path, audio_dur, self.max_dur, 0.0)

        audio_tensor = audio_tensor.squeeze(0)
        query_tensor = tensor2segment(pair["query_tensor"], self.max_dur)
        label = torch.LongTensor([1 if idx < len(self.positive_pairs) else -1])

        return audio_tensor, query_tensor, label

    def collate_fn(self, samples):
        """Collate a mini-batch of data."""
        audio_tensors, query_tensors, labels = zip(*samples)
        return audio_tensors + query_tensors, labels

    @property
    def sample_weights(self):
        """Sample weights to balance positive and negative data."""
        n_pos = len(self.positive_pairs)
        n_neg = len(self.negative_pairs)
        return [1 / n_pos] * n_pos + [1 / n_neg] * n_neg


def parse_rttm(rttm_path):
    """Parse audio and query pairs from *.rttm."""

    # e.g. "LEXEME sws2013_12345 ... 3.50 1.00 sws2013_dev_123 ..."
    pattern = re.compile(
        r"LEXEME\s+(sws2013_[0-9]+).*?([0-9]\.[0-9]+)\s+([0-9]\.[0-9]+)"
        r"\s+(sws2013_(dev|eval)_[0-9]+)"
    )

    query2audios = defaultdict(list)
    with open(rttm_path) as fd:
        for line in fd:
            match = pattern.match(line)
            if match is None:
                continue
            query2audios[match.group(4)].append(
                {
                    "audio": match.group(1),
                    "offset": float(match.group(2)),
                    "duration": float(match.group(3)),
                }
            )

    return query2audios


def parse_ecf(ecf_path):
    """Find audios from sws2013.ecf.xml."""

    root = ET.parse(str(ecf_path)).getroot()

    audio2dur = {}
    for excerpt in root.findall("excerpt"):
        audio_name = (
            excerpt.attrib["audio_filename"].replace("Audio/", "").replace(".wav", "")
        )
        duration = float(excerpt.attrib["dur"])
        audio2dur[audio_name] = duration

    return audio2dur


def find_queries(query_dir_path):
    """Find all queries under sws2013_dev & sws2013_eval."""

    # e.g. "sws2013_dev_123.wav" or "sws2013_dev_123_01.wav" -> "sws2013_dev_123"
    pattern = re.compile(r"(_[0-9]{2})?\.wav")

    query2tensors = defaultdict(list)
    for query_path in tqdm(
        list(query_dir_path.glob("*.wav")), ncols=0, desc="Load queries"
    ):
        query_name = pattern.sub("", query_path.name)
        wav_tensor, sample_rate = apply_effects_file(
            str(query_path), [["channels", "1"], ["rate", "16000"], ["norm"]]
        )
        trimmed, _ = apply_effects_tensor(
            wav_tensor,
            sample_rate,
            [
                ["vad", "-T", "0.25", "-p", "0.1"],
                ["reverse"],
                ["vad", "-T", "0.25", "-p", "0.1"],
                ["reverse"],
            ],
        )
        wav_tensor = trimmed if trimmed.size(1) >= (sample_rate * 0.5) else wav_tensor
        wav_tensor = wav_tensor.squeeze(0)
        query2tensors[query_name].append(wav_tensor)

    return query2tensors


def path2segment(filepath, src_dur, tgt_dur, offset):
    random_shift = random.uniform(0, src_dur - tgt_dur)
    audio_tensor, _ = apply_effects_file(
        str(filepath),
        [
            ["channels", "1"],
            ["rate", "16000"],
            ["norm"],
            ["pad", f"{tgt_dur}", f"{tgt_dur}"],
            [
                "trim",
                f"{tgt_dur + offset + random_shift}",
                f"{tgt_dur}",
            ],
        ],
    )
    return audio_tensor


def tensor2segment(tensor, tgt_dur, sample_rate=16000):
    src_dur = len(tensor) / sample_rate
    random_shift = random.uniform(0, src_dur - tgt_dur)
    audio_tensor, _ = apply_effects_tensor(
        tensor.unsqueeze(0),
        sample_rate,
        [
            ["pad", f"{tgt_dur}", f"{tgt_dur}"],
            [
                "trim",
                f"{tgt_dur + random_shift}",
                f"{tgt_dur}",
            ],
        ],
    )
    return audio_tensor.squeeze(0)



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/byol_a/expert.py ]
#   Synopsis     [ the BYOL-Audio wrapper ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import math
#-------------#
import torch
import torch.nn as nn
from torch.nn.utils.rnn import pad_sequence
#-------------#
import torchaudio
#-------------#
from .byol_a import load_yaml_config, PrecomputedNorm, AudioNTT2020


###################
# UPSTREAM EXPERT #
###################
class UpstreamExpert(nn.Module):
    """
    The BYOL-A wrapper
    """

    def __init__(self, ckpt, model_config=None, **kwargs):
        super(UpstreamExpert, self).__init__()

        if model_config is not None:
            print('[UpstreamExpert] - Using upstream expert config file from:', model_config) 
        else:
            model_config = './upstream/byol_a/config.yaml'
        config = load_yaml_config(model_config)

        # Preprocessor and normalizer.
        self.to_melspec = torchaudio.transforms.MelSpectrogram(
            sample_rate=config.sample_rate,
            n_fft=config.n_fft,
            win_length=config.win_length,
            hop_length=config.hop_length,
            n_mels=config.n_mels,
            f_min=config.f_min,
            f_max=config.f_max,
        )
        stats = [-5.4919195,  5.0389895] # provided by authors
        self.normalizer = PrecomputedNorm(stats)

        # Load pretrained weights.
        self.model = AudioNTT2020(d=config.feature_d)
        self.model.load_weight(ckpt, device='cpu')

        # attributes
        self.output_dim = config.feature_d
        self.max_input_length = config.shape[-1]

    # Interface
    def get_output_dim(self):
        return self.output_dim

    # Interface
    def get_downsample_rates(self, key: str) -> int:
        return 15344.655344655344 # computed by: len(wavs[0]) / len(features[0]) * self.max_input_length

    # forward in chunks
    def forward_in_chunks(self, features):
        outputs = []
        for i in range(0, features.size(1), self.max_input_length):
            subseq = features[:, i:i+self.max_input_length, :]
            if subseq.size(1) < self.max_input_length: break # break if the chunk is too small for the model to forward
            feats = self.model(subseq.permute(0, 2, 1).unsqueeze(1)) # features: (B, 1, F, T)
            outputs.append(feats.unsqueeze(1)) # (B, 1, D)
        outputs = torch.cat(outputs, dim=1) # (B, T, D)
        return outputs
    
    # Interface
    def forward(self, wavs):
        """
        Args:
            wavs:
                list of unpadded wavs [wav1, wav2, ...]
                each wav is in torch.FloatTensor with sample rate 16000
                and already put in the device assigned by command-line args

        Return:
            features:
                list of unpadded features [feat1, feat2, ...]
                each feat is in torch.FloatTensor and already
                put in the device assigned by command-line args
        """
        features = [self.normalizer((self.to_melspec(wav) + torch.finfo(torch.float).eps).log()).permute(1, 0) for wav in wavs] # features: (B, T, F)
        features = pad_sequence(features, batch_first=True)

        # forward the sequence in chunks then concat
        features = self.forward_in_chunks(features)
        return {
            "last_hidden_state": features,
            "hidden_states": [features],
        }



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ utils/audio.py ]
#   Synopsis     [ audio processing functions ]
#   Author       [ Andy T. Liu (Andi611) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
#   Reference 0  [ https://github.com/andi611/TTS-Tacotron-Pytorch ]
#   Reference 1  [ https://github.com/Alexander-H-Liu/End-to-end-ASR-Pytorch ]
#   Reference 2  [ https://groups.google.com/forum/#!msg/librosa/V4Z1HpTKn8Q/1-sMpjxjCSoJ ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import random
import numpy as np
#-------------#
import librosa
from scipy import signal
#-------------#
import matplotlib
matplotlib.use("Agg")
import matplotlib.pylab as plt
from matplotlib.colors import SymLogNorm
#-------------#
import warnings
warnings.filterwarnings("ignore")
# NOTE: there are warnings for MFCC extraction due to librosa's issue


##################
# AUDIO SETTINGS #
##################
sample_rate = 16000
"""
For feature == 'fbank' or 'mfcc'
"""
num_mels = 80 # int, dimension of feature
num_mfcc = 13 # int, number of MFCCs
window_size = 25 # int, window size for FFT (ms)
stride = 10 # int, window stride for FFT
"""
For feature == 'mel' or 'linear'
"""
num_freq = 1025
frame_length_ms = 50
frame_shift_ms = 12.5
preemphasis = 0.97
min_level_db = -100
ref_level_db = 20
hop_length = 250
griffin_lim_iters = 16
power = 1.5 # Power to raise magnitudes to prior to Griffin-Lim


#############################
# SPECTROGRAM UTILS FORWARD #
#############################
def _stft_parameters(sample_rate):
    n_fft = (num_freq - 1) * 2
    hop_length = int(frame_shift_ms / 1000 * sample_rate)
    win_length = int(frame_length_ms / 1000 * sample_rate)
    return n_fft, hop_length, win_length

def _linear_to_mel(spectrogram, sample_rate):
    _mel_basis = _build_mel_basis(sample_rate)
    return np.dot(_mel_basis, spectrogram)

def _build_mel_basis(sample_rate):
    n_fft = (num_freq - 1) * 2
    return librosa.filters.mel(sample_rate, n_fft, n_mels=num_mels)

def _preemphasis(x):
    return signal.lfilter([1, -preemphasis], [1], x)

def _amp_to_db(x):
    return 20 * np.log10(np.maximum(1e-5, x))

def _normalize(S):
    return np.clip((S - min_level_db) / -min_level_db, 0, 1)

def _stft(y, sr):
    n_fft, hop_length, win_length = _stft_parameters(sr)
    return librosa.stft(y=y, n_fft=n_fft, hop_length=hop_length, win_length=win_length)


#############################
# SPECTROGRAM UTILS BACKWARD #
#############################
def _denormalize(S):
    return (np.clip(S, 0, 1) * -min_level_db) + min_level_db

def _db_to_amp(x):
    return np.power(10.0, x * 0.05)

def inv_preemphasis(x):
    return signal.lfilter([1], [1, -preemphasis], x)

def _griffin_lim(S, sr):
    """
        librosa implementation of Griffin-Lim
        Based on https://github.com/librosa/librosa/issues/434
    """
    angles = np.exp(2j * np.pi * np.random.rand(*S.shape))
    S_complex = np.abs(S).astype(np.complex)
    y = _istft(S_complex * angles, sr)
    for _ in range(griffin_lim_iters):
        angles = np.exp(1j * np.angle(_stft(y ,sr)))
        y = _istft(S_complex * angles, sr)
    return y

def _istft(y, sr):
    _, hop_length, win_length = _stft_parameters(sr)
    return librosa.istft(y, hop_length=hop_length, win_length=win_length)


###################
# MEL SPECTROGRAM #
###################
"""
Compute the mel-scale spectrogram from the wav.
"""
def melspectrogram(y, sr):
    D = _stft(_preemphasis(y), sr)
    S = _amp_to_db(_linear_to_mel(np.abs(D), sr))
    return _normalize(S)


###############
# SPECTROGRAM #
###############
"""
Compute the linear-scale spectrogram from the wav.
"""
def spectrogram(y, sr):
    D = _stft(_preemphasis(y), sr)
    S = _amp_to_db(np.abs(D)) - ref_level_db
    return _normalize(S)


###################
# INV SPECTROGRAM #
###################
"""
Converts spectrogram to waveform using librosa
"""
def inv_spectrogram(spectrogram, sr=16000):
    S = _db_to_amp(_denormalize(spectrogram) + ref_level_db)  # Convert back to linear
    return inv_preemphasis(_griffin_lim(S ** power, sr))          # Reconstruct phase


###################
# EXTRACT FEATURE #
###################
# Acoustic Feature Extraction
# Parameters
#     - input file  : str, audio file path
#     - feature     : str, fbank or mfcc
#     - delta       : bool, append delta
#     - delta_delta : bool, append delta delta
#     - cmvn        : bool, apply CMVN on feature
#     - save_feature: str, if given, store feature to the path and return len(feature)
# Return
#     acoustic features with shape (time step, dim)
def extract_feature(input_file, feature='fbank', delta=False, delta_delta=False, cmvn=True, save_feature=None):
    y, sr = librosa.load(input_file, sr=sample_rate)

    if feature == 'fbank':
        ws = int(sr*0.001*window_size)
        st = int(sr*0.001*stride)
        feat = librosa.feature.melspectrogram(y=y, sr=sr, n_mels=num_mels,
                                              n_fft=ws, hop_length=st)
        feat = np.log(feat + 1e-6) # log-scaled
    elif feature == 'mfcc':
        ws = int(sr*0.001*window_size)
        st = int(sr*0.001*stride)
        feat = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=num_mfcc,
                                    n_fft=ws, hop_length=st)
        feat[0] = librosa.feature.rms(y, hop_length=st, frame_length=ws)
    elif feature == 'mel':
        # feat = melspectrogram(y, sr) # deprecated
        n_fft, hop_length, win_length = _stft_parameters(sr)
        feat = librosa.feature.melspectrogram(y=y, sr=sr, n_mels=num_mels,
                                              n_fft=n_fft, hop_length=hop_length, win_length=win_length)
        feat = np.log(feat + 1e-6) # log-scaled
    elif feature == 'linear':
        feat = spectrogram(y, sr)
    else:
        raise ValueError('Unsupported Acoustic Feature: ' + feature)

    # Apply delta
    feat = [feat]
    if delta and feature != 'linear':
        feat.append(librosa.feature.delta(feat[0]))

    if delta_delta and feature != 'linear':
        feat.append(librosa.feature.delta(feat[0], order=2))
    feat = np.concatenate(feat, axis=0)
    if feature == 'linear': assert(np.shape(feat)[0] == num_freq)

    if cmvn:
        feat = (feat - feat.mean(axis=1)[:,np.newaxis]) / (feat.std(axis=1)+1e-16)[:,np.newaxis]
    if save_feature is not None:
        tmp = np.swapaxes(feat, 0, 1).astype('float32')
        np.save(save_feature,tmp)
        return len(tmp)
    else:
        return np.swapaxes(feat, 0, 1).astype('float32')


#####################
# SAVE FIG TO NUMPY #
#####################
def _save_figure_to_numpy(fig):
    # save it to a numpy array.
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    return data.transpose(2, 0, 1) # (Channel, Height, Width)


#############################
# PLOT SPECTROGRAM TO NUMPY #
#############################
def plot_spectrogram_to_numpy(spectrogram):
    spectrogram = spectrogram.transpose(1, 0)
    fig, ax = plt.subplots(figsize=(12, 3))
    im = ax.imshow(spectrogram, aspect="auto", origin="lower",
                   interpolation='none')
    plt.colorbar(im, ax=ax)
    plt.xlabel("Frames")
    plt.ylabel("Channels")
    plt.tight_layout()

    fig.canvas.draw()
    data = _save_figure_to_numpy(fig)
    plt.close()
    return data


####################
# PLOT SPECTROGRAM #
####################
def plot_spectrogram(spec, path):
    spec = spec.transpose(1, 0) # (seq_len, feature_dim) -> (feature_dim, seq_len)
    plt.gcf().clear()
    plt.figure(figsize=(12, 3))
    plt.imshow(spec, aspect="auto", origin="lower")
    plt.colorbar()
    plt.tight_layout()
    plt.savefig(path, dpi=300, format="png")
    plt.close() 


####################
# PLOT EMBEDDING #
####################
def plot_embedding(spec, path):
    spec = spec.transpose(1, 0) # (seq_len, feature_dim) -> (feature_dim, seq_len)
    plt.gcf().clear()
    plt.figure(figsize=(12, 3))
    plt.pcolormesh(spec, norm=SymLogNorm(linthresh=1e-3))
    plt.colorbar()
    plt.tight_layout()
    plt.savefig(path, dpi=300, format="png")
    plt.close()


def plot_attention(attn, path):
    fig = plt.figure(figsize=(5, 5))
    plt.imshow(attn)
    plt.savefig(path, format='png')
    plt.close()



import torch
import torch.nn as nn
import torch.nn.functional as F


def get_downstream_model(input_dim, output_dim, config):
    model_cls = eval(config['select'])
    model_conf = config.get(config['select'], {})
    model = model_cls(input_dim, output_dim, **model_conf)
    return model


class FrameLevel(nn.Module):
    def __init__(self, input_dim, output_dim, hiddens=None, activation='ReLU', **kwargs):
        super().__init__()
        latest_dim = input_dim
        self.hiddens = []
        if hiddens is not None:
            for dim in hiddens:
                self.hiddens += [
                    nn.Linear(latest_dim, dim),
                    getattr(nn, activation)(),
                ]
                latest_dim = dim
        self.hiddens = nn.Sequential(*self.hiddens)
        self.linear = nn.Linear(latest_dim, output_dim)

    def forward(self, hidden_state, features_len=None):
        hidden_states = self.hiddens(hidden_state)
        logit = self.linear(hidden_state)

        return logit, features_len


class UtteranceLevel(nn.Module):
    def __init__(self,
        input_dim,
        output_dim,
        pooling='MeanPooling',
        activation='ReLU',
        pre_net=None,
        post_net={'select': 'FrameLevel'},
        **kwargs
    ):
        super().__init__()
        latest_dim = input_dim
        self.pre_net = get_downstream_model(latest_dim, latest_dim, pre_net) if isinstance(pre_net, dict) else None
        self.pooling = eval(pooling)(input_dim=latest_dim, activation=activation)
        self.post_net = get_downstream_model(latest_dim, output_dim, post_net)

    def forward(self, hidden_state, features_len=None):
        if self.pre_net is not None:
            hidden_state, features_len = self.pre_net(hidden_state, features_len)

        pooled, features_len = self.pooling(hidden_state, features_len)
        logit, features_len = self.post_net(pooled, features_len)

        return logit, features_len


class MeanPooling(nn.Module):

    def __init__(self, **kwargs):
        super(MeanPooling, self).__init__()

    def forward(self, feature_BxTxH, features_len, **kwargs):
        ''' 
        Arguments
            feature_BxTxH - [BxTxH]   Acoustic feature with shape 
            features_len  - [B] of feature length
        '''
        agg_vec_list = []
        for i in range(len(feature_BxTxH)):
            agg_vec = torch.mean(feature_BxTxH[i][:features_len[i]], dim=0)
            agg_vec_list.append(agg_vec)

        return torch.stack(agg_vec_list), torch.ones(len(feature_BxTxH)).long()


class AttentivePooling(nn.Module):
    ''' Attentive Pooling module incoporate attention mask'''

    def __init__(self, input_dim, activation, **kwargs):
        super(AttentivePooling, self).__init__()
        self.sap_layer = AttentivePoolingModule(input_dim, activation)

    def forward(self, feature_BxTxH, features_len):
        ''' 
        Arguments
            feature_BxTxH - [BxTxH]   Acoustic feature with shape 
            features_len  - [B] of feature length
        '''
        device = feature_BxTxH.device
        len_masks = torch.lt(torch.arange(features_len.max()).unsqueeze(0).to(device), features_len.unsqueeze(1))
        sap_vec, _ = self.sap_layer(feature_BxTxH, len_masks)

        return sap_vec, torch.ones(len(feature_BxTxH)).long()


class AttentivePoolingModule(nn.Module):
    """
    Implementation of Attentive Pooling 
    """
    def __init__(self, input_dim, activation='ReLU', **kwargs):
        super(AttentivePoolingModule, self).__init__()
        self.W_a = nn.Linear(input_dim, input_dim)
        self.W = nn.Linear(input_dim, 1)
        self.act_fn = getattr(nn, activation)()
        self.softmax = nn.functional.softmax

    def forward(self, batch_rep, att_mask):
        """
        input:
        batch_rep : size (B, T, H), B: batch size, T: sequence length, H: Hidden dimension
        
        attention_weight:
        att_w : size (B, T, 1)
        
        return:
        utter_rep: size (B, H)
        """
        att_logits = self.W(self.act_fn(self.W_a(batch_rep))).squeeze(-1)
        att_logits = att_mask + att_logits
        att_w = self.softmax(att_logits, dim=-1).unsqueeze(-1)
        utter_rep = torch.sum(batch_rep * att_w, dim=1)

        return utter_rep, att_w



import os
import torch
import random
import argparse
import transformers
from s3prl import hub
from packaging import version

SAMPLE_RATE = 16000
BATCH_SIZE = 8

parser = argparse.ArgumentParser()
parser.add_argument("--base", action="store_true")
parser.add_argument("--large", action="store_true")
parser.add_argument("--device", default="cuda")
args = parser.parse_args()

assert version.parse(transformers.__version__) <= version.parse(
    "4.9.0"
), "Newer version of transformers change the places for feature extraction."
assert args.base or args.large
s3prl_str = "wav2vec2_base_960" if args.base else "wav2vec2_large_ll60k"
huggingface_str = "wav2vec2_hug_base_960" if args.base else "wav2vec2_hug_large_ll60k"

s3prl = getattr(hub, s3prl_str)().to(args.device)
huggingface = getattr(hub, huggingface_str)().to(args.device)

if args.base:
    s3prl.wav_normalize = True
    s3prl.apply_padding_mask = False
s3prl.numpy_wav_normalize = True

s3prl.eval()
huggingface.eval()

wavs = [
    torch.randn(random.randint(SAMPLE_RATE * 1, SAMPLE_RATE * 15)).to(args.device)
    for _ in range(BATCH_SIZE)
]

with torch.no_grad():
    hiddens1 = s3prl(wavs)["hidden_states"]
    hiddens2 = huggingface(wavs)["hidden_states"]
    assert len(hiddens1) == len(hiddens2)

    diffs = []
    for idx, (hidden1, hidden2) in enumerate(zip(hiddens1, hiddens2)):
        diff = (hidden1 - hidden2).abs().max().item()
        print(f"hidden {idx} difference: {diff}")
        diffs.append(diff)

    print(f"Max difference: {torch.FloatTensor(diffs).max().item()}")



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ model.py ]
#   Synopsis     [ the linear model ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import torch
import torch.nn as nn
import torch.nn.functional as F

from argparse import Namespace
from s3prl.upstream.mockingjay.model import TransformerEncoder
#########
# MODEL #
#########

class Identity(nn.Module):
    def __init__(self, config, **kwargs):
        super(Identity, self).__init__()
        # simply take mean operator / no additional parameters

    def forward(self, feature, att_mask, head_mask, **kwargs):

        return [feature]

class Mean(nn.Module):

    def __init__(self, out_dim):
        super(Mean, self).__init__()
        self.act_fn = nn.Tanh()
        self.linear = nn.Linear(out_dim, out_dim)
        # simply take mean operator / no additional parameters

    def forward(self, feature, att_mask):

        ''' 
        Arguments
            feature - [BxTxD]   Acoustic feature with shape 
            att_mask   - [BxTx1]     Attention Mask logits
        '''
        feature=self.linear(self.act_fn(feature))
        agg_vec_list = []
        for i in range(len(feature)):
            if torch.nonzero(att_mask[i] < 0, as_tuple=False).size(0) == 0:
                length = len(feature[i])
            else:
                length = torch.nonzero(att_mask[i] < 0, as_tuple=False)[0] + 1
            agg_vec=torch.mean(feature[i][:length], dim=0)
            agg_vec_list.append(agg_vec)
        return torch.stack(agg_vec_list)

class SAP(nn.Module):
    ''' Self Attention Pooling module incoporate attention mask'''

    def __init__(self, out_dim):
        super(SAP, self).__init__()

        # Setup
        self.act_fn = nn.Tanh()
        self.sap_layer = SelfAttentionPooling(out_dim)
    
    def forward(self, feature, att_mask):

        ''' 
        Arguments
            feature - [BxTxD]   Acoustic feature with shape 
            att_mask   - [BxTx1]     Attention Mask logits
        '''
        #Encode
        feature = self.act_fn(feature)
        sap_vec = self.sap_layer(feature, att_mask)

        return sap_vec

class SelfAttentionPooling(nn.Module):
    """
    Implementation of SelfAttentionPooling 
    Original Paper: Self-Attention Encoding and Pooling for Speaker Recognition
    https://arxiv.org/pdf/2008.01077v1.pdf
    """
    def __init__(self, input_dim):
        super(SelfAttentionPooling, self).__init__()
        self.W = nn.Linear(input_dim, 1)
    def forward(self, batch_rep, att_mask):
        """
        input:
        batch_rep : size (N, T, H), N: batch size, T: sequence length, H: Hidden dimension
        
        attention_weight:
        att_w : size (N, T, 1)
        
        return:
        utter_rep: size (N, H)
        """
        seq_len = batch_rep.shape[1]
        softmax = nn.functional.softmax
        att_logits = self.W(batch_rep).squeeze(-1)
        att_logits = att_mask + att_logits
        att_w = softmax(att_logits, dim=-1).unsqueeze(-1)
        utter_rep = torch.sum(batch_rep * att_w, dim=1)

        return utter_rep

class Model(nn.Module):
    def __init__(self, input_dim, agg_module, output_class_num, config):
        super(Model, self).__init__()
        
        # agg_module: current support [ "SAP", "Mean" ]
        # init attributes
        self.agg_method = eval(agg_module)(input_dim)
        self.linear = nn.Linear(input_dim, output_class_num)
        
        # two standard transformer encoder layer
        self.model= eval(config['module'])(config=Namespace(**config['hparams']),)
        self.head_mask = [None] * config['hparams']['num_hidden_layers']         


    def forward(self, features, att_mask):
        features = self.model(features,att_mask[:,None,None], head_mask=self.head_mask, output_all_encoded_layers=False)
        utterance_vector = self.agg_method(features[0], att_mask)
        predicted = self.linear(utterance_vector)
        
        return predicted
        # Use LogSoftmax since self.criterion combines nn.LogSoftmax() and nn.NLLLoss()



import torch.nn as nn

from s3prl.upstream.wav2vec2.wav2vec2_model import (
    Wav2Vec2Config,
    TransformerEncoder,
    TransformerSentenceEncoderLayer,
)

args = {
    "input_feat": 80,
    "encoder_embed_dim": 768,
    "encoder_layers": 12,
    "dropout": 0.1,
    "activation_dropout": 0,
    "dropout_input": 0.1,
    "attention_dropout": 0.1,
    "encoder_layerdrop": 0.05,
    "conv_pos": 128,
    "conv_pos_groups": 16,
    "encoder_ffn_embed_dim": 3072,
    "encoder_attention_heads": 12,
    "activation_fn": "gelu",
    "layer_norm_first": False,
}

class Decoar2(nn.Module):
    def __init__(self):
        """
        input_size: an int indicating the input feature size, e.g., 80 for Mel.
        hidden_size: an int indicating the RNN hidden size.
        num_layers: an int indicating the number of RNN layers.
        dropout: a float indicating the RNN dropout rate.
        residual: a bool indicating whether to apply residual connections.
        """
        super(Decoar2, self).__init__()
        config = Wav2Vec2Config()
        for arg_name, arg_val in args.items():
            setattr(config, arg_name, arg_val)

        self.post_extract_proj = nn.Linear(config.input_feat, config.encoder_embed_dim)
        self.dropout_input = nn.Dropout(config.dropout_input)
        self.encoder = TransformerEncoder(config)

    def forward(self, features, padding_mask=None):
        features = self.post_extract_proj(features)

        x = self.dropout_input(features)

        if padding_mask is not None:
            extra = padding_mask.size(1) % features.size(1)
            if extra > 0:
                padding_mask = padding_mask[:, :-extra]
            padding_mask = padding_mask.view(padding_mask.size(0), features.size(1), -1)
            padding_mask = padding_mask.all(-1)

        x, layer_results = self.encoder(x, padding_mask=padding_mask, layer=None)

        return x, layer_results



import torch
import torch.nn as nn
import torch.nn.functional as F


class Model(nn.Module):
    def __init__(self, input_dim, bottleneck_dim, hidden_dim, **kwargs):
        super(Model, self).__init__()

        self.connector = nn.Linear(input_dim, bottleneck_dim)
        self.fc1 = nn.Linear(bottleneck_dim, hidden_dim)
        self.attention_linear = nn.Linear(hidden_dim, 1)

    def forward(self, features):
        # transforming
        hiddens = F.relu(self.connector(features))
        hiddens = torch.tanh(self.fc1(hiddens))

        # attentive pooling
        attention_weights = F.softmax(self.attention_linear(hiddens), dim=1)
        embeds = torch.sum(hiddens * attention_weights, dim=1)

        return embeds



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ find_best_epoch.py ]
#   Synopsis     [ Script to read and find best epoch (used with batch training mode only) ]
#   Author       [ Wen-Chin Huang (https://github.com/unilight) ]
#   Copyright    [ Copyright(c), Toda Lab, Nagoya University, Japan ]
"""*********************************************************************************************"""


import argparse
from io import open
import os

srcspks = ["SEF1", "SEF2", "SEM1", "SEM2"]
task1_trgspks = ["TEF1", "TEF2", "TEM1", "TEM2"]
task2_trgspks = ["TFF1", "TFM1", "TGF1", "TGM1", "TMF1", "TMM1"]


def get_parser():
    parser = argparse.ArgumentParser(
        description="Extract results.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--upstream", type=str, required=True, help="upstream")
    parser.add_argument("--task", type=str, required=True, choices=["task1", "task2"], help="task")
    parser.add_argument("--tag", type=str, required=True, help="tag")
    parser.add_argument("--vocoder", type=str, required=True, help="vocoder name")
    parser.add_argument("--expdir", type=str, default="result/downstream", help="expdir")
    parser.add_argument("--start_epoch", default=4000, type=int)
    parser.add_argument("--end_epoch", default=10000, type=int)
    parser.add_argument("--step_epoch", default=1000, type=int)
    parser.add_argument(
        "--out",
        "-O",
        type=str,
        help="The output filename. " "If omitted, then output to sys.stdout",
    )
    return parser


def grep(filepath, query):
    lines = []
    with open(filepath, "r") as f:
        for line in f:
            if query in line:
                lines.append(line.rstrip())
    return lines


if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()

    epochs = list(range(args.start_epoch, args.end_epoch+args.step_epoch, args.step_epoch))
    bests = []
    trgspks = task1_trgspks if args.task == "task1" else task2_trgspks

    for trgspk in trgspks:
        scores = []
        for ep in epochs:
            # the expdir name pattern is consistent with the batch training script
            log_file = os.path.join(args.expdir, f"a2o_vc_vcc2020_{args.tag}_{trgspk}_{args.upstream}", str(ep), f"{args.vocoder}_wav", "obj.log")
            if args.task == "task1":
                result = grep(log_file, "Mean")[0].split("Mean MCD, f0RMSE, f0CORR, DDUR, CER, WER, accept rate: ")[1].split(" ")
            elif args.task == "task2":
                result = grep(log_file, "Mean")[0].split("Mean CER, WER, accept rate: ")[1].split(" ")
            scores.append(result)
        
        # task 1: choose by MCD; task 2: choose by CER
        best = min(scores, key=lambda x: float(x[0]))
        bests.append(best)

    if args.task == "task1":
        avg = [f"{(sum([float(best[i]) for best in bests]) / 4.0):.2f}" for i in range(7)]
    elif args.task == "task2":
        avg = [f"{(sum([float(best[i]) for best in bests]) / 6.0):.2f}" for i in range(3)]
    print("Best result:"+" ".join(avg))



import os
from librosa.util import find_files
import random

roots = {"Voxceleb1":"/home/pohan/data/librispeech/vox1_train_verifi/wav", "Voxceleb2":"/home/pohan/data/librispeech/vox2_dev/wav"}

def collect_speaker_ids(roots, speaker_num):
    
    all_speaker=[ ]
    for key in list(roots.keys()):
        all_speaker.extend([f.path for f in os.scandir(roots[key]) if f.is_dir()])
    
    ids = [[speaker.split("/")[-3],speaker.split("/")[-1]] for speaker in all_speaker]

    vox1 = []
    vox2 = []
    
    for id in ids:
        if id[0] == roots["Voxceleb1"].split("/")[-2]:
            vox1.append(id[1])
        if id[0] == roots["Voxceleb2"].split("/")[-2]:
            vox2.append(id[1])

    dev_speaker = random.sample(vox1, k=speaker_num)
    vox1_train = [ids for ids in vox1 if ids not in dev_speaker]
    
    train_speaker = []

    train_speaker.extend(vox1_train)
    train_speaker.extend(vox2)

    return train_speaker, dev_speaker

def construct_dev_speaker_id_txt(dev_speakers,dev_txt_name):
    f = open(dev_txt_name, "w")
    for dev in dev_speakers:
        f.write(dev)
        f.write("\n")
    f.close()
    return

def sample_wavs_and_dump_txt(root,dev_ids, numbers, meta_data_name):
    
    wav_list = []
    count_positive = 0
    for _ in range(numbers):
        prob = random.random()
        if (prob > 0.5):
            dev_id_pair = random.sample(dev_ids, 2)

            # sample 2 wavs from different speaker
            sample1 = "/".join(random.choice(find_files(os.path.join(root,dev_id_pair[0]))).split("/")[-3:])
            sample2 = "/".join(random.choice(find_files(os.path.join(root,dev_id_pair[1]))).split("/")[-3:])

            label = "0"

            wav_list.append(" ".join([label, sample1, sample2]))
            
        else:
            dev_id_pair = random.sample(dev_ids, 1)
            
            # sample 2 wavs from same speaker
            sample1 = "/".join(random.choice(find_files(os.path.join(root,dev_id_pair[0]))).split("/")[-3:])
            sample2 = "/".join(random.choice(find_files(os.path.join(root,dev_id_pair[0]))).split("/")[-3:])

            label = "1"
            count_positive +=1

            wav_list.append(" ".join([label, sample1, sample2]))
    
    f = open(meta_data_name,"w")
    for data in wav_list:
        f.write(data+"\n")
    f.close()

    return wav_list


if __name__ == "__main__":
    train_speakers, dev_speakers = collect_speaker_ids(roots, 51)
    construct_dev_speaker_id_txt(dev_speakers, "dev_meta_data/dev_speaker_ids.txt")
    wav_list = sample_wavs_and_dump_txt(roots["Voxceleb1"], dev_speakers, 4000, "dev_meta_data/dev_meta_data.txt")




import torch
import torch.nn as nn
import torchaudio
from torchaudio.compliance import kaldi


############
# CONSTANT #
############
WINDOW_TYPE = 'hamming'
SAMPLE_RATE = 16000


class CMVN(torch.jit.ScriptModule):
    
    __constants__ = ["mode", "dim", "eps"]

    def __init__(self, mode="global", dim=2, eps=1e-10):
        # `torchaudio.load()` loads audio with shape [channel, feature_dim, time]
        # so perform normalization on dim=2 by default
        super(CMVN, self).__init__()

        if mode != "global":
            raise NotImplementedError(
                "Only support global mean variance normalization.")

        self.mode = mode
        self.dim = dim
        self.eps = eps

    @torch.jit.script_method
    def forward(self, x):
        if self.mode == "global":
            return (x - x.mean(self.dim, keepdim=True)) \
                   / (self.eps + x.std(self.dim, keepdim=True))

    def extra_repr(self):
        return "mode={}, dim={}, eps={}".format(self.mode, self.dim, self.eps)


class FeatureExtractor(nn.Module):
    ''' Feature extractor, transforming file path to Mel spectrogram '''
    def __init__(self, mode="fbank", num_mel_bins=80, decode_wav=False, 
                 apply_cmvn=True, **kwargs):
        super(FeatureExtractor, self).__init__()
        # ToDo: Other surface representation
        assert mode=="fbank", "Only Mel-spectrogram implemented"
        self.mode = mode
        self.extract_fn = kaldi.fbank
        self.apply_cmvn = apply_cmvn
        if self.apply_cmvn:
            self.cmvn = CMVN()
        self.num_mel_bins = num_mel_bins
        self.kwargs = kwargs
        self.decode_wav = decode_wav
        if self.decode_wav:
            # HACK: sox cannot deal with wav with incorrect file length
            torchaudio.set_audio_backend('soundfile')

    def _load_file(self, filepath):
        if self.decode_wav:
            waveform, sample_rate = torchaudio.load_wav(filepath)
        else:
            waveform, sample_rate = torchaudio.load(filepath)
        return waveform, sample_rate

    def forward(self, waveform):
        y = self.extract_fn(waveform,
                            num_mel_bins=self.num_mel_bins,
                            sample_frequency=SAMPLE_RATE,
                            window_type = WINDOW_TYPE,
                            **self.kwargs)
        # CMVN
        if self.apply_cmvn:
            y = y.transpose(0, 1).unsqueeze(0) # TxD -> 1xDxT
            y = self.cmvn(y)
            y = y.squeeze(0).transpose(0,1) #  1xDxT -> TxD

        #Low Frame Rate: Take every 2 frame
        out = y[::2, :]

        return out

    def extra_repr(self):
        return "mode={}, num_mel_bins={}".format(self.mode, self.num_mel_bins)

    def create_msg(self):
        ''' List msg for verbose function '''
        msg = 'Audio spec.| Audio feat. = {}\t\t| feat. dim = {}\t| CMVN = {}'\
                        .format(self.mode, self.num_mel_bins, self.apply_cmvn)
        return [msg]


def create_transform():
    return FeatureExtractor()


# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ dataset.py ]
#   Synopsis     [ the phone dataset ]
#   Author       [ S3PRL, Xuankai Chang ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""
# Adaptive_SpecAugment Author: ShampooWang, cornliu

###############
# IMPORTATION #
###############
import os
import random
#-------------#
import pandas as pd
#-------------#
import torch
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data.dataset import Dataset
#-------------#
import torchaudio


DEFAULT_TIME_WARP_MODE = "bicubic"

class SpecAug(torch.nn.Module):
    """SpecAug"""
    def __init__(
        self,
        apply_time_warp=True,
        time_warp_window=5,
        time_warp_mode="bicubic",
        apply_freq_mask=True,
        freq_mask_width_range=(0,20),
        num_freq_mask=2,
        apply_time_mask=True,
        time_mask_width_range=(0,100),
        num_time_mask=2,
        adaptive_number_ratio = 0.04,
        adaptive_size_ratio = 0.04,
        max_n_time_masks = 20,
        adaptive=False
    ):
        assert any([apply_time_warp, apply_freq_mask, apply_time_mask])

        super(SpecAug, self).__init__()
        self.apply_time_warp = apply_time_warp
        self.apply_freq_mask = apply_freq_mask
        self.apply_time_mask = apply_time_mask

        if apply_time_warp:
            self.time_warp = TimeWarp(window=time_warp_window, mode=time_warp_mode)
        else:
            self.time_warp = None

        if apply_freq_mask:
            self.freq_mask = MaskAlongAxis(
                dim="freq",
                mask_width_range=freq_mask_width_range,
                num_mask=num_freq_mask,
            )
        else:
            self.freq_mask = None

        if apply_time_mask:
            self.time_mask = MaskAlongAxis(
                dim="time",
                mask_width_range=time_mask_width_range,
                num_mask=num_time_mask,
                adaptive=adaptive,
                adaptive_number_ratio=adaptive_number_ratio,
                adaptive_size_ratio=adaptive_size_ratio,
                max_n_time_masks=max_n_time_masks
            )
        else:
            self.time_mask = None

    def apply_specaug(self, x, x_lengths=None):
        if self.time_warp is not None:
            x, x_lengths = self.time_warp(x, x_lengths)
        if self.freq_mask is not None:
            x, x_lengths = self.freq_mask(x, x_lengths)
        if self.time_mask is not None:
            x, x_lengths = self.time_mask(x, x_lengths)
        return x, x_lengths

    def forward(self, xs, x_lengths=None):
        """Forward SpecAug.
        
        Args:
            xs: list of features [(T, D)] x batchsize
            x_lengths: length of features
        Return:
            list of features[(T, D)] x batchsize
        """
        assert len(xs[0].size()) == 2
        if x_lengths is None:
            x_lengths = torch.LongTensor([x.size(0) for x in xs])

        batchsize, max_len, dim = len(x_lengths), torch.max(x_lengths).item(), xs[0].size(1)
        xs_pad = xs[0].new_zeros((batchsize, max_len, dim))
        for i, x in enumerate(xs):
            xs_pad[i, :x_lengths[i]] = x

        xs_pad, _ = self.apply_specaug(xs_pad, x_lengths)

        xs = [xs_pad[i, :xs[i].size(0)] for i in range(batchsize)]
        return xs, x_lengths


class TimeWarp(torch.nn.Module):
    """Time warping using torch.interpolate.
    Args:
        window: time warp parameter
        mode: Interpolate mode
    """

    def __init__(self, window=80, mode=DEFAULT_TIME_WARP_MODE):
        super().__init__()
        self.window = window
        self.mode = mode

    def extra_repr(self):
        return f"window={self.window}, mode={self.mode}"

    def time_warp(self, x):
        org_size = x.size()
        if x.dim() == 3:
            # x: (Batch, Time, Freq) -> (Batch, 1, Time, Freq)
            x = x[:, None]

        t = x.shape[2]
        if t - self.window <= self.window:
            return x.view(*org_size)

        center = torch.randint(self.window, t - self.window, (1,))[0]
        warped = torch.randint(center - self.window, center + self.window, (1,))[0] + 1

        # left: (Batch, Channel, warped, Freq)
        # right: (Batch, Channel, time - warped, Freq)
        left = torch.nn.functional.interpolate(
            x[:, :, :center], (warped, x.shape[3]), mode=self.mode, align_corners=False
        )
        right = torch.nn.functional.interpolate(
            x[:, :, center:], (t - warped, x.shape[3]), mode=self.mode, align_corners=False
        )

        if x.requires_grad:
            x = torch.cat([left, right], dim=-2)
        else:
            x[:, :, :warped] = left
            x[:, :, warped:] = right

        return x.view(*org_size)

    def forward(self, x, x_lengths=None):
        """Forward function.
        Args:
            x: (Batch, Time, Freq)
            x_lengths: (Batch,)
        """
        ys = x.new_zeros(x.size())
        for i in range(x.size(0)):
            _y = self.time_warp(
                x[i][None, : x_lengths[i]],
            )[0]
            ys[i, : x_lengths[i]] = _y

        return ys, x_lengths


class MaskAlongAxis(torch.nn.Module):
    def __init__(
        self,
        mask_width_range=(0, 30),
        num_mask=2,
        dim="time",
        replace_with_zero=True,
        adaptive_number_ratio = 0.04,
        adaptive_size_ratio = 0.04,
        max_n_time_masks = 20,
        adaptive = False
    ):
        if isinstance(mask_width_range, int):
            mask_width_range = (0, mask_width_range)
        if len(mask_width_range) != 2:
            raise TypeError(
                f"mask_width_range must be a tuple of int and int values: "
                f"{mask_width_range}",
            )

        assert mask_width_range[1] > mask_width_range[0]
        if isinstance(dim, str):
            if dim == "time":
                dim = 1
            elif dim == "freq":
                dim = 2
            else:
                raise ValueError("dim must be int, 'time' or 'freq'")
        if dim == 1:
            self.mask_axis = "time"
        elif dim == 2:
            self.mask_axis = "freq"
        else:
            self.mask_axis = "unknown"

        super().__init__()
        self.mask_width_range = mask_width_range
        self.num_mask = num_mask
        self.dim = dim
        self.replace_with_zero = replace_with_zero

        ###############################################
        self.adaptive = adaptive
        self.adaptive_number_ratio = adaptive_number_ratio
        self.adaptive_size_ratio = adaptive_size_ratio
        self.max_n_time_masks = max_n_time_masks
        ###############################################

    def mask_along_axis(self, spec, spec_lengths):
        org_size = spec.size()
        if spec.dim() == 4:
            # spec: (Batch, Channel, Length, Freq) -> (Batch * Channel, Length, Freq)
            spec = spec.view(-1, spec.size(2), spec.size(3))

        B = spec.shape[0]

        # D = Length or Freq
        D = spec.shape[self.dim]
        T = self.mask_width_range[1]
        num_mask = self.num_mask
        
        
        if self.dim == 1 & self.adaptive :
          if self.adaptive_number_ratio > 0:
            num_mask = min(int(self.adaptive_number_ratio * D), self.max_n_time_masks)
          if self.adaptive_size_ratio > 0:
            T = min(self.mask_width_range[1], int(self.adaptive_size_ratio * D))

        # mask_length: (B, num_mask, 1)
        mask_length = torch.randint(
            self.mask_width_range[0],
            T,
            (B, num_mask),
            device=spec.device,
        ).unsqueeze(2)

        # mask_pos: (B, num_mask, 1)
        mask_pos = torch.randint(
            0, max(1, D - mask_length.max()), (B, num_mask), device=spec.device
        ).unsqueeze(2)

        # aran: (1, 1, D)
        aran = torch.arange(D, device=spec.device)[None, None, :]
        # mask: (Batch, num_mask, D)
        mask = (mask_pos <= aran) * (aran < (mask_pos + mask_length))
        # Multiply masks: (Batch, num_mask, D) -> (Batch, D)
        mask = mask.any(dim=1)
        if self.dim == 1:
            # mask: (Batch, Length, 1)
            mask = mask.unsqueeze(2)
        elif self.dim == 2:
            # mask: (Batch, 1, Freq)
            mask = mask.unsqueeze(1)

        if self.replace_with_zero:
            value = 0.0
        else:
            value = spec.mean()

        if spec.requires_grad:
            spec = spec.masked_fill(mask, value)
        else:
            spec = spec.masked_fill_(mask, value)
        spec = spec.view(*org_size)
        return spec, spec_lengths

    def forward(self, spec, spec_lengths=None):
        """Forward function.
        Args:
            spec: (Batch, Length, Freq)
        """

        return self.mask_along_axis(
            spec,
            spec_lengths,
        )



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/cpc/hubconf.py ]
#   Synopsis     [ the cpc torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import os

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def cpc_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    assert os.path.isfile(ckpt)
    return _UpstreamExpert(ckpt, *args, **kwargs)


def cpc_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from URL
        ckpt (str): URL
    """
    return cpc_local(_urls_to_filepaths(ckpt), *args, **kwargs)


def modified_cpc(refresh=False, *args, **kwargs):
    """
    The model from official repository
    """
    kwargs[
        "ckpt"
    ] = "https://dl.fbaipublicfiles.com/librilight/CPC_checkpoints/60k_epoch4-d0f474de.pt"
    return cpc_url(refresh=refresh, *args, **kwargs)



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ pretrain/apc/pretrain_expert.py ]
#   Synopsis     [ the apc pretrain expert ]
#   Author       [ Andy T. Liu (https://github.com/andi611) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import yaml
import copy
#-------------#
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
#-------------#
from pretrain.apc.dataset import ApcAudioDataset
from utility.audio import plot_spectrogram_to_numpy


####################
# UPSTREAM WRAPPER #
####################
class UpstreamPretrainExpert(nn.Module):
    """
    The APC pretrain expert
    """

    def __init__(self, datarc, upstream_config, device='cuda', multi_gpu=False, **kwargs):
        super(UpstreamPretrainExpert, self).__init__()

        self.datarc = datarc
        self.device = device
        self.multi_gpu = multi_gpu

        if type(upstream_config) == str:
            self.upstream_config = yaml.load(open(upstream_config, 'r'), Loader=yaml.FullLoader)
            print('[UpstreamPretrainExpert] - Using upstream config from:', upstream_config)
        elif type(upstream_config) == dict:
            self.upstream_config = upstream_config
            print('[UpstreamPretrainExpert] - Using upstream config from the previous experiment.')
        else:
            raise ValueError
        
        preprocessor = self._init_model()
        self._get_train_dataloader(preprocessor)

        if self.multi_gpu:
            self.model = torch.nn.DataParallel(self.model)
            print('[UpstreamPretrainExpert] - Multi-GPU training Enabled: ' + str(torch.cuda.device_count()))
        print('[UpstreamPretrainExpert] - Number of parameters: ' + str(sum(p.numel() for p in self.model.parameters() if p.requires_grad)))

    def _init_model(self):
        from upstream.apc.audio import create_transform
        from upstream.apc.apc import APC

        try:
            print('[UpstreamPretrainExpert] - Using the apc preprocessor, on-the-fly feature preprocessing')
            preprocessor, feat_dim = create_transform(copy.deepcopy(self.upstream_config['data']['audio']))
        except:
            raise NotImplementedError('Our upstream wrapper currently does not support other feature extracters, see: `s3prl/upstream/apc/expert.py`')
        
        print('[UpstreamPretrainExpert] - Initializing model...')
        self.model = APC(feat_dim, **self.upstream_config["model"]["paras"])
        self.n_future = self.upstream_config["task"]["n_future"]
        self.loss = torch.nn.L1Loss()
        return preprocessor

    def _get_train_dataloader(self, preprocessor):
        dataset = ApcAudioDataset(preprocessor,
                                  self.upstream_config['task'],
                                  self.datarc['train_batch_size'],
                                  **self.datarc)
        self.dataloader = DataLoader(dataset, batch_size=1, # for bucketing
                                     shuffle=True, num_workers=self.datarc['num_workers'],
                                     drop_last=False, pin_memory=True, collate_fn=dataset.collate_fn)

    # Interface
    def load_model(self, init_ckpt):
        assert 'model' in init_ckpt
        if self.multi_gpu:
            self.model.module.load_state_dict(init_ckpt['model'])
        else:
            self.model.load_state_dict(init_ckpt['model'])

    # Interface
    def loss_to_device(self):
        self.loss.to(self.device)

    # Interface
    def add_state_to_save(self, all_states):
        all_states['config'] = self.upstream_config
        all_states['model'] = self.model.state_dict() if not self.multi_gpu else \
                                 self.model.module.state_dict()
        all_states['Upstream_Config'] = self.upstream_config
        return all_states

    # Interface
    def get_train_dataloader(self):
        return self.dataloader

    # Interface
    def forward(self, data, records={}, global_step=0, log_step=1000, **kwargs):
        """
        Args:
            data:
                [spec_masked, pos_enc, mask_label, attn_mask, spec_target]
            
            records:
                defaultdict(list), by appending contents into records,
                these contents can be averaged and logged on Tensorboard
                later by self.log_records every log_step

        Return:
            loss        
        """

        audio_feat, audio_len = data[0], data[1]
        audio_feat = audio_feat.to(self.device)
        
        # APC input = shifted target
        audio_len = [l-self.n_future for l in audio_len]
        pred_spec, _ = self.model(audio_feat[:,:-self.n_future,:], audio_len, testing=False)
        loss = self.loss(pred_spec, audio_feat[:,self.n_future:,:])

        if global_step % log_step == 0:
            spec_list = [pred_spec, audio_feat]
            name_list = ['pred_spec', 'true_spec']
            
            for i in range(len(spec_list)):
                spec = plot_spectrogram_to_numpy(spec_list[i][0].data.cpu().numpy())
                records[name_list[i]] = spec
            
        return loss, records

    # interface
    def on_before_zero_grad(self):
        pass
    
    # interface
    def log_records(self, records, logger, prefix, global_step, **kwargs):
        """
        Args:
            records:
                defaultdict(list), contents already appended

            logger:
                Tensorboard SummaryWriter
                please use f'{prefix}your_content_name' as key name
                to log your customized contents

            prefix:
                used to indicate downstream and train/test on Tensorboard
                eg. 'phone/train-'

            global_step:
                global_step in runner, which is helpful for Tensorboard logging
        """
        for key, values in records.items():
            logger.add_image(
                f'{prefix}{key}',
                values,
                global_step=global_step
            )


import torch
import logging
import argparse
from copy import deepcopy
from dataclasses import is_dataclass, dataclass

from s3prl.util.pseudo_data import get_pseudo_wavs
from s3prl.util.download import _urls_to_filepaths

logger = logging.getLogger(__name__)


def load_fairseq_ckpt(source: str, **override):
    from omegaconf import OmegaConf
    from fairseq.checkpoint_utils import load_checkpoint_to_cpu

    source = str(source)
    if source.startswith("http"):
        fairseq_path = _urls_to_filepaths(source)
    else:
        fairseq_path = source

    state = load_checkpoint_to_cpu(fairseq_path, arg_overrides=override)
    cfg = OmegaConf.to_container(state["cfg"])

    assert type(cfg) == dict
    return state, cfg


def merge_with_parent(dc: dataclass, cfg: dict):

    assert is_dataclass(dc)
    assert type(cfg) == dict
    cfg = deepcopy(cfg)

    def fix_cfg(cfg):
        target_keys = set(dc.__dataclass_fields__.keys())
        for k in list(cfg.keys()):
            if k not in target_keys:
                del cfg[k]

    fix_cfg(cfg)
    assert len(cfg) > 0
    return dc(**cfg)


def extract_hidden_states(model):
    model.eval()
    with torch.no_grad():
        return model(get_pseudo_wavs())["hidden_states"]


def are_same_models(model1, model2):
    hs1 = extract_hidden_states(model1)
    hs2 = extract_hidden_states(model2)
    for h1, h2 in zip(hs1, hs2):
        assert torch.allclose(h1, h2)


def models_all_close(*models):
    assert len(models) > 1
    for model in models[1:]:
        are_same_models(models[0], model)



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/mockingjay/hubconf.py ]
#   Synopsis     [ the mockingjay torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import os
import torch

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def mockingjay_local(ckpt, options_config=None, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
        feature_selection (int): -1 (default, the last layer) or an int in range(0, max_layer_num)
    """
    assert os.path.isfile(ckpt)
    if options_config is not None:
        assert os.path.isfile(options_config)
    return _UpstreamExpert(ckpt, options_config, *args, **kwargs)


def mockingjay_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from URL
        ckpt (str): URL
    """
    return mockingjay_local(_urls_to_filepaths(ckpt, refresh=refresh), *args, **kwargs)


def mockingjay(refresh=False, *args, **kwargs):
    """
    The default model
        refresh (bool): whether to download ckpt/config again if existed
    """
    return mockingjay_origin(refresh=refresh, *args, **kwargs)


###########
# ALIASES #
###########


def mockingjay_origin(refresh=False, *args, **kwargs):
    """
    The mockingjay large model on 360hr, with Lel as input and Linear as target
        refresh (bool): whether to download ckpt/config again if existed
    """
    return mockingjay_logMelLinearLarge_T_AdamW_b32_500k_360hr_drop1(
        refresh=refresh, *args, **kwargs
    )


def mockingjay_100hr(refresh=False, *args, **kwargs):
    """
    The mockingjay base model on 100hr
        refresh (bool): whether to download ckpt/config again if existed
    """
    return mockingjay_logMelBase_T_AdamW_b32_200k_100hr(
        refresh=refresh, *args, **kwargs
    )


def mockingjay_960hr(refresh=False, *args, **kwargs):
    """
    The mockingjay base model on 960hr
        refresh (bool): whether to download ckpt/config again if existed
    """
    return mockingjay_logMelBase_T_AdamW_b32_1m_960hr_drop1(
        refresh=refresh, *args, **kwargs
    )


##########
# 100 HR #
##########


def mockingjay_logMelBase_T_AdamW_b32_200k_100hr(refresh=False, *args, **kwargs):
    """
    Feature: 80-dim log Mel
    Alteration: time
    Optimizer: AdamW
    Batch size: 32
    Total steps: 200k
    Unlabled Speech: 100hr
    """
    kwargs["ckpt"] = "https://www.dropbox.com/s/luorglf8mdg67l2/states-200000.ckpt?dl=1"
    return mockingjay_url(refresh=refresh, *args, **kwargs)


##########
# 360 HR #
##########


def mockingjay_logMelLinearLarge_T_AdamW_b32_500k_360hr_drop1(
    refresh=False, *args, **kwargs
):
    """
    Feature: 80-dim log Mel (input) / 201-dim Linear (target)
    Alteration: time
    Optimizer: AdamW
    Batch size: 32
    Total steps: 500k
    Unlabled Speech: 360hr
    """
    kwargs["ckpt"] = "https://www.dropbox.com/s/zwsfa6w2iy2cc68/states-500000.ckpt?dl=1"
    return mockingjay_url(refresh=refresh, *args, **kwargs)


##########
# 960 HR #
##########


def mockingjay_logMelBase_T_AdamW_b32_1m_960hr(refresh=False, *args, **kwargs):
    """
    Feature: 80-dim log Mel
    Alteration: time
    Optimizer: AdamW
    Batch size: 32
    Total steps: 1M
    Unlabled Speech: 960hr
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/jzx0xggk663jev6/states-1000000.ckpt?dl=1"
    return mockingjay_url(refresh=refresh, *args, **kwargs)


def mockingjay_logMelBase_T_AdamW_b32_1m_960hr_drop1(refresh=False, *args, **kwargs):
    """
    Feature: 80-dim log Mel
    Alteration: time
    Optimizer: AdamW
    Batch size: 32
    Total steps: 1M
    Unlabled Speech: 960hr
    Differences: Dropout of 0.1 (instead of 0.3)
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/7f9z6dzc7oix6qv/states-1000000.ckpt?dl=1"
    return mockingjay_url(refresh=refresh, *args, **kwargs)


def mockingjay_logMelBase_T_AdamW_b32_1m_960hr_seq3k(refresh=False, *args, **kwargs):
    """
    Feature: 80-dim log Mel
    Alteration: time
    Optimizer: AdamW
    Batch size: 32
    Total steps: 1M
    Unlabled Speech: 960hr
    Differences: sequence length of 3k (instead of 1.5k)
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/qnnvdrai2tfmjmh/states-1000000.ckpt?dl=1"
    return mockingjay_url(refresh=refresh, *args, **kwargs)


import os
import math
import torch
import random

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence

from .model import Model
from .dataset import AtisDataset
import pandas as pd
from collections import Counter


class DownstreamExpert(nn.Module):
    """
    Used to handle downstream-specific operations
    eg. downstream forward, metric computation, contents to log
    """

    def __init__(self, upstream_dim, downstream_expert, **kwargs):
        super(DownstreamExpert, self).__init__()
        self.upstream_dim = upstream_dim
        self.datarc = downstream_expert['datarc']
        self.modelrc = downstream_expert['modelrc']
        
        self.get_dataset()

        self.train_dataset = AtisDataset(self.train_df, self.base_path, self.Sy_intent, 'train')
        self.dev_dataset = AtisDataset(self.valid_df, self.base_path, self.Sy_intent, 'dev')
        self.test_dataset = AtisDataset(self.test_df, self.base_path, self.Sy_intent, 'test')

        self.connector = nn.Linear(upstream_dim, self.modelrc['input_dim'])
        self.model = Model(input_dim=self.modelrc['input_dim'], agg_module=self.modelrc['agg_module'],output_dim=sum(self.values_per_slot), config=self.modelrc)
        self.objective = nn.CrossEntropyLoss()

    def get_dataset(self):
        self.base_path = self.datarc['file_path']
        train_df = pd.read_csv(os.path.join(self.base_path, "nlu_iob", "iob.train"), sep='\t', header=None)
        valid_df = pd.read_csv(os.path.join(self.base_path, "nlu_iob", "iob.dev"), sep='\t', header=None)
        test_df = pd.read_csv(os.path.join(self.base_path, "nlu_iob", "iob.test"), sep='\t', header=None)

        train_dict = {"id": {}, "intent": {}}
        valid_dict = {"id": {}, "intent": {}}
        test_dict = {"id": {}, "intent": {}}

        
        for dc, df, type_name in [(train_dict, train_df, 'train'), (valid_dict, valid_df, 'dev'), (test_dict, test_df, 'test')]:
            n = 0
            for i in range(len(df)):
                if(os.path.exists( os.path.join(self.base_path, type_name, df[0][i].split()[0]+'.wav')) & (df[0][i].split()[0] not in dc['id'].values())):
                    dc['id'][n] = df[0][i].split()[0]
                    dc['intent'][n] = df[1][i].split()[-1]
                    n += 1
                # else:
                #     print(os.path.join(self.base_path, type_name, df[0][i].split()[0]+'.wav'))
            print(type_name, ': ', n+1)    

        train_df = pd.DataFrame(data=train_dict)
        valid_df = pd.DataFrame(data=valid_dict)
        test_df = pd.DataFrame(data=test_dict)


        Sy_intent = {"intent": {}}
        values_per_slot = []
        
        for slot in ["intent"]:
            slot_values = Counter(train_df[slot]) + Counter(valid_df[slot]) + Counter(test_df[slot])
            for index, value in enumerate(slot_values):
                Sy_intent[slot][value] = index
            values_per_slot.append(len(slot_values))
        self.values_per_slot = values_per_slot
        self.Sy_intent = Sy_intent
        self.train_df = train_df
        self.valid_df = valid_df
        self.test_df = test_df

    def _get_train_dataloader(self, dataset):
        return DataLoader(
            dataset, batch_size=self.datarc['train_batch_size'],
            shuffle=True, num_workers=self.datarc['num_workers'],
            collate_fn=dataset.collate_fn
        )

    def _get_eval_dataloader(self, dataset):
        return DataLoader(
            dataset, batch_size=self.datarc['eval_batch_size'],
            shuffle=False, num_workers=self.datarc['num_workers'],
            collate_fn=dataset.collate_fn
        )

    def get_train_dataloader(self):
        return self._get_train_dataloader(self.train_dataset)

    def get_dev_dataloader(self):
        return self._get_eval_dataloader(self.dev_dataset)

    def get_test_dataloader(self):
        return self._get_eval_dataloader(self.test_dataset)

    def get_dataloader(self, mode):
        return eval(f'self.get_{mode}_dataloader')()

    # Interface
    def forward(self, mode, features, labels, records=None, **kwargs):

        features_pad = pad_sequence(features, batch_first=True)
        
        attention_mask = [torch.ones((feature.shape[0])) for feature in features] 

        attention_mask_pad = pad_sequence(attention_mask,batch_first=True)

        attention_mask_pad = (1.0 - attention_mask_pad) * -100000.0

        features_pad = self.connector(features_pad)
        intent_logits = self.model(features_pad, attention_mask_pad.cuda())

        intent_loss = 0
        start_index = 0
        predicted_intent = []
        #labels = torch.LongTensor(labels).to(features_pad.device)
        
        labels = torch.stack(labels).to(features_pad.device)
        for slot in range(len(self.values_per_slot)):
            end_index = start_index + self.values_per_slot[slot]
            subset = intent_logits[:, start_index:end_index]

            intent_loss += self.objective(subset, labels[:, slot])
            predicted_intent.append(subset.max(1)[1])
            start_index = end_index

        predicted_intent = torch.stack(predicted_intent, dim=1)
        records['acc'] += (predicted_intent == labels).prod(1).view(-1).cpu().float().tolist()
        records['intent_loss'].append(intent_loss.item())

        return intent_loss

    # interface
    def log_records(self, mode, records, logger, global_step, **kwargs):
        for key, values in records.items():
            average = torch.FloatTensor(values).mean().item()
            logger.add_scalar(
                f'atis/{mode}-{key}',
                average,
                global_step=global_step
            )



import os

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def mos_wav2vec2_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    assert os.path.isfile(ckpt)
    kwargs["upstream"] = "wav2vec2"
    return _UpstreamExpert(ckpt, *args, **kwargs)


def mos_wav2vec2_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from URL
        ckpt (str): URL
    """
    return mos_wav2vec2_local(_urls_to_filepaths(ckpt), *args, **kwargs)


def mos_wav2vec2(refresh=False, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/s9zpouk5svu1a4l/wav2vec2-dev-SRCC-best.ckpt?dl=1"
    return mos_wav2vec2_url(refresh=refresh, *args, **kwargs)


def mos_tera_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    assert os.path.isfile(ckpt)
    kwargs["upstream"] = "tera"
    return _UpstreamExpert(ckpt, *args, **kwargs)


def mos_tera_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from URL
        ckpt (str): URL
    """
    return mos_tera_local(_urls_to_filepaths(ckpt), *args, **kwargs)


def mos_tera(refresh=False, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/w4jk5bujaoosk69/tera-dev-SRCC-best.ckpt?dl=1"
    return mos_tera_url(refresh=refresh, *args, **kwargs)


def mos_apc_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    assert os.path.isfile(ckpt)
    kwargs["upstream"] = "apc"
    return _UpstreamExpert(ckpt, *args, **kwargs)


def mos_apc_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from URL
        ckpt (str): URL
    """
    return mos_apc_local(_urls_to_filepaths(ckpt), *args, **kwargs)


def mos_apc(refresh=False, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/ulng31as15hsvz1/apc-dev-SRCC-best.ckpt?dl=1"
    return mos_apc_url(refresh=refresh, *args, **kwargs)



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ pretrain/apc/dataset.py ]
#   Synopsis     [ the dataset that applies the apc preprocessing on audio ]
#   Author       [ Andy T. Liu (https://github.com/andi611) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import os
import numpy as np
#-------------#
import torch
from torch.nn.utils.rnn import pad_sequence
import torchaudio
#-------------#
from pretrain.bucket_dataset import FeatDataset


class ApcAudioDataset(FeatDataset):
    
    def __init__(self, extracter, task_config, bucket_size, file_path, sets, 
                 max_timestep=0, libri_root=None, **kwargs):
        super(ApcAudioDataset, self).__init__(extracter, task_config, bucket_size, file_path, sets, 
                                                   max_timestep, libri_root, **kwargs)

    def _load_feat(self, feat_path):
        if self.libri_root is None:
            return torch.FloatTensor(np.load(os.path.join(self.root, feat_path)))
        else:
            wav, _ = torchaudio.load(os.path.join(self.libri_root, feat_path))
            feat = self.extracter(wav)
            return feat

    def __getitem__(self, index):
        # Load acoustic feature and pad
        x_batch = [self._sample(self._load_feat(x_file)) for x_file in self.X[index]]
        x_len = [len(x_b) for x_b in x_batch]
        x_pad_batch = pad_sequence(x_batch, batch_first=True)
        return x_pad_batch, x_len



from pathlib import Path
import os

import random
import torch
from torch.utils.data.dataset import Dataset
from torchaudio.sox_effects import apply_effects_file
from itertools import accumulate


class VCC18SegmentalDataset(Dataset):
    def __init__(self, dataframe, base_path, idtable = '', valid = False):
        self.base_path = Path(base_path)
        self.dataframe = dataframe
        self.segments_durations = 1
        if Path.is_file(idtable):
            self.idtable = torch.load(idtable)
            for i, judge_i in enumerate(self.dataframe['JUDGE']):
                self.dataframe['JUDGE'][i] = self.idtable[judge_i]

        elif not valid:
            self.gen_idtable(idtable)

    def __len__(self):
        return len(self.dataframe)

    def __getitem__(self, idx):
        wav_name, mean, mos, judge_id = self.dataframe.loc[idx]
        wav_path = self.base_path / "Converted_speech_of_submitted_systems" / wav_name
        wav, _ = apply_effects_file(
            str(wav_path),
            [
                ["channels", "1"],
                ["rate", "16000"],
                ["norm"],
            ],
        )

        wav = wav.view(-1)
        wav_segments = unfold_segments(wav, self.segments_durations)
        system_name = wav_name[:3] + wav_name[-8:-4]

        return wav_segments, mean, system_name, mos, judge_id

    def collate_fn(self, samples):
        wavs_segments, means, system_names, moss, judge_ids = zip(*samples)
        flattened_wavs_segments = [
            wav_segment
            for wav_segments in wavs_segments
            for wav_segment in wav_segments
        ]
        wav_segments_lengths = [len(wav_segments) for wav_segments in wavs_segments]
        prefix_sums = list(accumulate(wav_segments_lengths, initial=0))
        segment_judge_ids = []
        for i in range(len(prefix_sums)-1):
            segment_judge_ids.extend([judge_ids[i]] * (prefix_sums[i+1]-prefix_sums[i]))
        
        return (
            torch.stack(flattened_wavs_segments),
            prefix_sums,
            torch.FloatTensor(means),
            system_names,
            torch.FloatTensor(moss), 
            torch.LongTensor(segment_judge_ids)
        )
    
    def gen_idtable(self, idtable_path):
        if idtable_path == '':
            idtable_path = './idtable.pkl'
        self.idtable = {}
        count = 0
        for i, judge_i in enumerate(self.dataframe['JUDGE']):
            if judge_i not in self.idtable.keys():
                self.idtable[judge_i] = count
                count += 1
                self.dataframe['JUDGE'][i] = self.idtable[judge_i]
            else:
                self.dataframe['JUDGE'][i] = self.idtable[judge_i]
        torch.save(self.idtable, idtable_path)


class VCC16SegmentalDataset(Dataset):
    def __init__(self, wav_list, base_path):
        self.wav_dir = Path(base_path)
        self.wav_list = wav_list
        self.segments_durations = 1

    def __len__(self):
        return len(self.wav_list)

    def __getitem__(self, idx):
        wav_name = self.wav_list[idx]
        wav_path = self.wav_dir / wav_name
        wav, _ = apply_effects_file(
            str(wav_path),
            [
                ["channels", "1"],
                ["rate", "16000"],
                ["norm"],
            ],
        )

        wav = wav.view(-1)
        wav_segments = unfold_segments(wav, self.segments_durations)
        system_name = wav_name.name.split("_")[0]

        return wav_segments, system_name

    def collate_fn(self, samples):
        wavs_segments, system_names = zip(*samples)
        flattened_wavs_segments = [
            wav_segment
            for wav_segments in wavs_segments
            for wav_segment in wav_segments
        ]
        wav_segments_lengths = [len(wav_segments) for wav_segments in wavs_segments]
        prefix_sums = list(accumulate(wav_segments_lengths, initial=0))

        return (
            torch.stack(flattened_wavs_segments),
            prefix_sums,
            None,
            system_names,
            None, 
            None,
        )


def unfold_segments(tensor, tgt_duration, sample_rate=16000):
    seg_lengths = int(tgt_duration * sample_rate)
    src_lengths = len(tensor)
    step = seg_lengths // 2
    tgt_lengths = (
        seg_lengths if src_lengths <= seg_lengths else (src_lengths // step + 1) * step
    )

    pad_lengths = tgt_lengths - src_lengths
    padded_tensor = torch.cat([tensor, torch.zeros(pad_lengths)])
    segments = padded_tensor.unfold(0, seg_lengths, step).unbind(0)

    return segments



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ observe_input.py ]
#   Synopsis     [ generates visualizations of the model's input ]
#   Author       [ Ting-Wei Liu (Andi611) ]
#   Copyright    [ Copyleft(c), NTUEE, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import os
import copy
import random
import numpy as np
#-------------#
import torch
import torchaudio
#-------------#
import matplotlib
import matplotlib.pyplot as plt
#-------------#
plt.switch_backend('agg')
seed = 679
random.seed(seed)
np.random.seed(seed)
torch.manual_seed(seed)


########
# PATH #
########
utt = 'train-clean-100/1594/135914/1594-135914-0032.flac'
libri_dir = '/media/andi611/1TBSSD/LibriSpeech/'
out_dir = './result/visualization/'


def plot_x(x, name='x', xlabel='Frames'):
    x = x.transpose(1, 0)
    fig, ax = plt.subplots(figsize=(10, 3))
    im = ax.imshow(x, aspect='auto', origin='lower',
                   interpolation='none')
    plt.colorbar(im, ax=ax)
    plt.xlabel(xlabel)
    plt.ylabel('Channels')
    plt.tight_layout()
    plt.margins(0,0)
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())

    fig.canvas.draw()
    fig.savefig(os.path.join(out_dir, name + '.png'), bbox_inches='tight', pad_inches = 0)


def starts_to_intervals(starts, consecutive):
    tiled = starts.expand(consecutive, starts.size(0)).T
    offset = torch.arange(consecutive).expand_as(tiled)
    intervals = tiled + offset
    return intervals.view(-1)


########
# MAIN #
########
def main():
    
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)

    # plot original
    extracter = torch.hub.load('s3prl/s3prl', 'mel')
    wav, _ = torchaudio.load(os.path.join(libri_dir, utt))
    wavs = [wav]
    x = extracter(wavs)[0].squeeze()
    plot_x(x, name='x', xlabel='A) Original fMLLR feature')

    # to torch tensor
    x = torch.FloatTensor(x)
    x_all = copy.deepcopy(x)

    # time masking
    mask_consecutive_min = 7
    mask_consecutive_max = 7
    mask_proportion = 0.15
    mask_allow_overlap = True

    mask_consecutive = random.randint(mask_consecutive_min, mask_consecutive_max)
    valid_start_max = max(x.size(0) - mask_consecutive - 1, 0) # compute max valid start point for a consecutive mask
    proportion = round(x.size(0) * mask_proportion / mask_consecutive)
    if mask_allow_overlap:
        # draw `proportion` samples from the range (0, valid_index_range) and without replacement
        chosen_starts = torch.randperm(valid_start_max + 1)[:proportion]
    else:
        mask_bucket_size = round(mask_consecutive * mask_bucket_ratio)
        rand_start = random.randint(0, min(mask_consecutive, valid_start_max))
        valid_starts = torch.arange(rand_start, valid_start_max + 1, mask_bucket_size)
        chosen_starts = valid_starts[torch.randperm(len(valid_starts))[:proportion]]
    chosen_intervals = starts_to_intervals(chosen_starts, mask_consecutive)
    
    # mask to zero
    x_time_zero = copy.deepcopy(x)
    x_time_zero[chosen_intervals, :] = 0
    x_all[chosen_intervals, :] = 0
    plot_x(x_time_zero.data.cpu().numpy(), name='x_time_zero', xlabel='B) Mask contiguous segments to zero along temporal axis')

    # replace to random frames
    random_starts = torch.randperm(valid_start_max + 1)[:proportion]
    random_intervals = starts_to_intervals(random_starts, mask_consecutive)
    x_time_replace = copy.deepcopy(x)
    x_time_replace[chosen_intervals, :] = x_time_replace[random_intervals, :]
    plot_x(x_time_replace.data.cpu().numpy(), name='x_time_replace', xlabel='C) Replace contiguous segments with random segments')

    # frequency masking
    mask_frequency = 16
    rand_bandwidth = mask_frequency #random.randint(0, mask_frequency)
    chosen_starts = torch.randperm(x.size(1) - rand_bandwidth)[:1]
    chosen_intervals = starts_to_intervals(chosen_starts, rand_bandwidth)
    x_freq = copy.deepcopy(x)
    x_freq[:, chosen_intervals] = 0
    x_all[:, chosen_intervals] = 0
    plot_x(x_freq.data.cpu().numpy(), name='x_freq', xlabel='D) Mask contiguous segments to zero along channel axis')

    # noise augmentation
    noise_sampler = torch.distributions.Normal(0, 0.2)
    x_noise = copy.deepcopy(x)
    x_noise += noise_sampler.sample(x_noise.shape)
    x_all += noise_sampler.sample(x_all.shape)
    plot_x(x_noise.data.cpu().numpy(), name='x_noise', xlabel='E) Apply sampled Gaussian noise to magnitude')

    # time + freq + noise
    plot_x(x_all.data.cpu().numpy(), name='x_all', xlabel='F) Combining the alterations in B), D), and E)')


if __name__ == '__main__':
	main()


import os

import torch

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def decoar2_custom(ckpt: str, refresh=False, *args, **kwargs):
    if ckpt.startswith("http"):
        ckpt = _urls_to_filepaths(ckpt, refresh=refresh)

    return _UpstreamExpert(ckpt, *args, **kwargs)


def decoar2_local(*args, **kwargs):
    """
        The model from local ckpt
            ckpt (str): PATH
            feature_selection (str): 'c' (default) or 'z'
    """
    return decoar2_custom(*args, **kwargs)


def decoar2_url(*args, **kwargs):
    """
        The model from URL
            ckpt (str): URL
    """
    return decoar2_custom(*args, **kwargs)


def decoar2(*args, refresh=False, **kwargs):
    """
        The apc standard model on 360hr
            refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/checkpoint_decoar2.pt"
    return decoar2_url(*args, refresh=refresh, **kwargs)



import re
import xml.etree.ElementTree as ET
from pathlib import Path

from torch.utils.data.dataset import Dataset
from torchaudio.sox_effects import apply_effects_file


class SWS2013Testset(Dataset):
    """SWS 2013 testset."""

    def __init__(self, split, **kwargs):
        assert split in ["dev", "eval"]

        scoring_root = Path(kwargs["sws2013_scoring_root"])
        audio_names = parse_ecf(scoring_root / f"sws2013_{split}" / "sws2013.ecf.xml")
        query_names = parse_tlist(
            scoring_root / f"sws2013_{split}" / f"sws2013_{split}.tlist.xml"
        )

        self.dataset_root = Path(kwargs["sws2013_root"])
        self.split = split
        self.n_queries = len(query_names)
        self.n_docs = len(audio_names)
        self.data = query_names + audio_names

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        audio_name = self.data[idx]
        audio_path = (
            (self.dataset_root / f"{self.split}_queries" / audio_name)
            if idx < self.n_queries
            else (self.dataset_root / "Audio" / audio_name)
        )
        audio_path = audio_path.with_suffix(".wav")
        wav, _ = apply_effects_file(
            str(audio_path),
            [
                ["channels", "1"],
                ["rate", "16000"],
                ["norm"],
                ["vad", "-T", "0.25", "-p", "0.1"],
                ["reverse"],
                ["vad", "-T", "0.25", "-p", "0.1"],
                ["reverse"],
                ["pad", "0", "3"],
            ],
        )
        segments = wav.squeeze(0).unfold(0, 48000, 12000).unbind(0)
        return segments, len(segments), audio_name

    def collate_fn(self, samples):
        """Collate a mini-batch of data."""
        segments, lengths, audio_names = zip(*samples)
        segments = [seg for segs in segments for seg in segs]
        return segments, (lengths, audio_names)


def parse_ecf(ecf_path):
    """Find audio paths from sws2013.ecf.xml."""

    root = ET.parse(str(ecf_path)).getroot()

    audio_names = []
    for excerpt in root.findall("excerpt"):
        audio_name = (
            excerpt.attrib["audio_filename"].replace("Audio/", "").replace(".wav", "")
        )
        audio_names.append(audio_name)

    return audio_names


def parse_tlist(tlist_path):
    """Find audio paths from sws2013_eval.tlist.xml."""

    root = ET.parse(str(tlist_path)).getroot()

    audio_names = []
    for term in root.findall("term"):
        audio_names.append(term.attrib["termid"])

    return audio_names



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ dataset.py ]
#   Synopsis     [ LibriMix speech enhancement dataset ]
#   Author       [ Zili Huang ]
#   Copyright    [ Copyright(c), Johns Hopkins University ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import os
import random

import numpy as np

import torch
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data.dataset import Dataset

import librosa

SAMPLE_RATE = 16000

class SeparationDataset(Dataset):
    def __init__(
        self,
        data_dir,
        rate=16000,
        src=['mix_clean'],
        tgt=['s1', 's2'],
        n_fft=512,
        hop_length=320,
        win_length=512,
        window='hann', 
        center=True,
    ):
        super(SeparationDataset, self).__init__()
        """
        Args:
            data_dir (str):
                prepared data directory

            rate (int):
                audio sample rate

            src and tgt (list(str)):
                the input and desired output.
                LibriMix offeres different options for the users. For
                clean source separation, src=['mix_clean'] and tgt=['s1', 's2'].
                Please see https://github.com/JorisCos/LibriMix for details

            n_fft (int):
                length of the windowed signal after padding with zeros.

            hop_length (int):
                number of audio samples between adjacent STFT columns.

            win_length (int):
                length of window for each frame

            window (str):
                type of window function, only support Hann window now

            center (bool):
                whether to pad input on both sides so that the
                t-th frame is centered at time t * hop_length

            The STFT related parameters are the same as librosa.
        """

        self.data_dir = data_dir
        self.rate = rate
        self.src = src
        self.tgt = tgt
        self.n_fft = n_fft
        self.hop_length = hop_length
        self.win_length = win_length
        self.window = window
        self.center = center
        self.n_srcs = len(self.tgt)

        assert len(self.src) == 1 and len(self.tgt) == 1

        # mix_clean (utterances only) mix_both (utterances + noise) mix_single (1 utterance + noise)
        cond_list = ["s1", "s2", "noise", "mix_clean", "mix_both", "mix_single", "noisy", "clean"]

        # create the mapping from utterances to the audio paths
        # reco2path[utt][cond] is the path for utterance utt with condition cond
        reco2path = {}
        for cond in src + tgt:
            assert cond in cond_list
            assert os.path.exists("{}/{}/wav.scp".format(self.data_dir, cond))
            with open("{}/{}/wav.scp".format(self.data_dir, cond), 'r') as fh:
                content = fh.readlines()
            for line in content:
                line = line.strip('\n')
                uttname, path = line.split()
                if uttname not in reco2path:
                    reco2path[uttname] = {}
                reco2path[uttname][cond] = path
        self.reco2path = reco2path

        self.recolist = list(self.reco2path.keys())
        self.recolist.sort()

    def __len__(self):
        return len(self.recolist)

    def __getitem__(self, i):
        reco = self.recolist[i]
        src_path = self.reco2path[reco][self.src[0]]
        src_samp, rate = librosa.load(src_path, sr=SAMPLE_RATE)
        assert rate == self.rate
        src_feat = np.transpose(librosa.stft(src_samp, 
            n_fft=self.n_fft,
            hop_length = self.hop_length,
            win_length = self.win_length,
            window = self.window,
            center = self.center))

        tgt_samp_list, tgt_feat_list = [], []
        for j in range(self.n_srcs):
            tgt_path = self.reco2path[reco][self.tgt[j]]
            tgt_samp, rate = librosa.load(tgt_path, sr=SAMPLE_RATE)
            assert rate == self.rate
            tgt_feat = np.transpose(librosa.stft(tgt_samp, 
                n_fft=self.n_fft,
                hop_length = self.hop_length,
                win_length = self.win_length,
                window = self.window,
                center = self.center))
            tgt_samp_list.append(tgt_samp)
            tgt_feat_list.append(tgt_feat)
        """
        reco (str):
            name of the utterance

        src_sample (ndarray):
            audio samples for the source [T, ]

        src_feat (ndarray):
            the STFT feature map for the source with shape [T1, D]

        tgt_samp_list (list(ndarray)):
            list of audio samples for the targets

        tgt_feat_list (list(ndarray)):
            list of STFT feature map for the targets
        """
        return reco, src_samp, src_feat, tgt_samp_list, tgt_feat_list

    def collate_fn(self, batch):
        sorted_batch = sorted(batch, key=lambda x: -x[1].shape[0])
        bs = len(sorted_batch)
        uttname_list = [sorted_batch[i][0] for i in range(bs)]

        # Store the magnitude, phase for the mixture in source_attr
        source_attr = {}
        mix_magnitude_list = [torch.from_numpy(np.abs(sorted_batch[i][2])) for i in range(bs)]
        mix_phase_list = [torch.from_numpy(np.angle(sorted_batch[i][2])) for i in range(bs)]
        mix_stft_list = [torch.from_numpy(sorted_batch[i][2]) for i in range(bs)]
        mix_magnitude = pad_sequence(mix_magnitude_list, batch_first=True)
        mix_phase = pad_sequence(mix_phase_list, batch_first=True)
        mix_stft = pad_sequence(mix_stft_list, batch_first=True)
        source_attr["magnitude"] = mix_magnitude
        source_attr["phase"] = mix_phase
        source_attr["stft"] = mix_stft

        target_attr = {}
        target_attr["magnitude"] = []
        target_attr["phase"] = []
        for j in range(self.n_srcs):
            tgt_magnitude_list = [torch.from_numpy(np.abs(sorted_batch[i][4][j])) for i in range(bs)]
            tgt_phase_list = [torch.from_numpy(np.angle(sorted_batch[i][4][j])) for i in range(bs)]
            tgt_magnitude = pad_sequence(tgt_magnitude_list, batch_first=True)
            tgt_phase = pad_sequence(tgt_phase_list, batch_first=True)
            target_attr["magnitude"].append(tgt_magnitude)
            target_attr["phase"].append(tgt_phase)

        wav_length = torch.from_numpy(np.array([len(sorted_batch[i][1]) for i in range(bs)]))
        source_wav_list = [torch.from_numpy(sorted_batch[i][1]) for i in range(bs)]
        source_wav = pad_sequence(source_wav_list, batch_first=True)
        target_wav_list = []
        for j in range(self.n_srcs):
            target_wav_list.append(pad_sequence([torch.from_numpy(sorted_batch[i][3][j]) for i in range(bs)], batch_first=True))

        feat_length = torch.from_numpy(np.array([stft.size(0) for stft in mix_stft_list]))
        """
        source_wav_list (list(tensor)):
            list of audio samples for the source

        uttname_list (list(str)):
            list of utterance names

        source_attr (dict):
            dictionary containing magnitude and phase information for the sources

        source_wav (tensor):
            padded version of source_wav_list, with size [bs, max_T]

        target_attr (dict):
            dictionary containing magnitude and phase information for the targets

        feat_length (tensor):
            length of the STFT feature for each utterance

        wav_length (tensor):
            number of samples in each utterance
        """
        return source_wav_list, uttname_list, source_attr, source_wav, target_attr, target_wav_list, feat_length, wav_length



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/byol_a/hubconf.py ]
#   Synopsis     [ the BYOL-A torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""

import os

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert

# FIXME (Leo): These are wrong according to the original author

# def byol_a_local(ckpt, model_config=None, *args, **kwargs):
#     """
#         The model from local ckpt
#             ckpt (str): PATH
#     """
#     assert os.path.isfile(ckpt)
#     if model_config is not None:
#         assert os.path.isfile(model_config)
#     return _UpstreamExpert(ckpt, model_config, *args, **kwargs)


# def byol_a_url(ckpt, refresh=False, *args, **kwargs):
#     """
#         The model from URL
#             ckpt (str): URL
#     """
#     return byol_a_local(_urls_to_filepaths(ckpt, refresh=refresh), *args, **kwargs)


# def byol_a(refresh=False, *args, **kwargs):
#     """
#         The default model
#             refresh (bool): whether to download ckpt/config again if existed
#     """
#     return byol_a_2048(refresh=refresh, *args, **kwargs)


# def byol_a_2048(refresh=False, *args, **kwargs):
#     """
#             refresh (bool): whether to download ckpt/config again if existed
#     """
#     kwargs['ckpt'] = 'https://github.com/nttcslab/byol-a/raw/master/pretrained_weights/AudioNTT2020-BYOLA-64x96d2048.pth'
#     return byol_a_url(refresh=refresh, *args, **kwargs)


# def byol_a_1024(refresh=False, *args, **kwargs):
#     """
#     refresh (bool): whether to download ckpt/config again if existed
#     """
#     kwargs[
        "ckpt"
    ] = "https://github.com/nttcslab/byol-a/raw/master/pretrained_weights/AudioNTT2020-BYOLA-64x96d1024.pth"
#     return byol_a_url(refresh=refresh, *args, **kwargs)


# def byol_a_512(refresh=False, *args, **kwargs):
#     """
#     refresh (bool): whether to download ckpt/config again if existed
#     """
#     kwargs[
        "ckpt"
    ] = "https://github.com/nttcslab/byol-a/raw/master/pretrained_weights/AudioNTT2020-BYOLA-64x96d512.pth"
#     return byol_a_url(refresh=refresh, *args, **kwargs)




import os
import math
import torch
import random
import pandas as pd
from pathlib import Path
from collections import Counter

import torch
import torch.nn as nn
from torch.utils.data import DataLoader, DistributedSampler
from torch.distributed import is_initialized
from torch.nn.utils.rnn import pad_sequence

from ..model import *
from .dataset import FluentCommandsDataset


class DownstreamExpert(nn.Module):
    """
    Used to handle downstream-specific operations
    eg. downstream forward, metric computation, contents to log
    """

    def __init__(self, upstream_dim, downstream_expert, expdir, **kwargs):
        super(DownstreamExpert, self).__init__()
        self.upstream_dim = upstream_dim
        self.datarc = downstream_expert['datarc']
        self.modelrc = downstream_expert['modelrc']

        self.get_dataset()

        self.train_dataset = FluentCommandsDataset(self.train_df, self.base_path, self.Sy_intent)
        self.dev_dataset = FluentCommandsDataset(self.valid_df, self.base_path, self.Sy_intent)
        self.test_dataset = FluentCommandsDataset(self.test_df, self.base_path, self.Sy_intent)

        model_cls = eval(self.modelrc['select'])
        model_conf = self.modelrc.get(self.modelrc['select'], {})
        self.projector = nn.Linear(upstream_dim, self.modelrc['projector_dim'])
        self.model = model_cls(
            input_dim = self.modelrc['projector_dim'],
            output_dim = sum(self.values_per_slot),
            **model_conf,
        )
        self.objective = nn.CrossEntropyLoss()
        self.expdir = expdir
        self.register_buffer('best_score', torch.zeros(1))

    def get_dataset(self):
        self.base_path = self.datarc['file_path']
        train_df = pd.read_csv(os.path.join(self.base_path, "data", "train_data.csv"))
        valid_df = pd.read_csv(os.path.join(self.base_path, "data", "valid_data.csv"))
        test_df = pd.read_csv(os.path.join(self.base_path, "data", "test_data.csv"))

        Sy_intent = {"action": {}, "object": {}, "location": {}}

        values_per_slot = []
        for slot in ["action", "object", "location"]:
            slot_values = Counter(train_df[slot])
            for index, value in enumerate(slot_values):
                Sy_intent[slot][value] = index
                Sy_intent[slot][index] = value
            values_per_slot.append(len(slot_values))
        self.values_per_slot = values_per_slot
        self.Sy_intent = Sy_intent
        self.train_df = train_df
        self.valid_df = valid_df
        self.test_df = test_df

    def _get_train_dataloader(self, dataset):
        sampler = DistributedSampler(dataset) if is_initialized() else None
        return DataLoader(
            dataset, batch_size=self.datarc['train_batch_size'],
            shuffle=(sampler is None), sampler=sampler,
            num_workers=self.datarc['num_workers'],
            collate_fn=dataset.collate_fn
        )

    def _get_eval_dataloader(self, dataset):
        return DataLoader(
            dataset, batch_size=self.datarc['eval_batch_size'],
            shuffle=False, num_workers=self.datarc['num_workers'],
            collate_fn=dataset.collate_fn
        )

    def get_train_dataloader(self):
        return self._get_train_dataloader(self.train_dataset)

    def get_dev_dataloader(self):
        return self._get_eval_dataloader(self.dev_dataset)

    def get_test_dataloader(self):
        return self._get_eval_dataloader(self.test_dataset)

    # Interface
    def get_dataloader(self, mode):
        return eval(f'self.get_{mode}_dataloader')()

    # Interface
    def forward(self, mode, features, labels, filenames, records, **kwargs):
        labels = [torch.LongTensor(label) for label in labels]
        features_len = torch.IntTensor([len(feat) for feat in features]).to(device=features[0].device)
        features = pad_sequence(features, batch_first=True)

        features = self.projector(features)
        intent_logits, _ = self.model(features, features_len)

        intent_loss = 0
        start_index = 0
        predicted_intent = []
        
        labels = torch.stack(labels).to(features.device)
        for slot in range(len(self.values_per_slot)):
            end_index = start_index + self.values_per_slot[slot]
            subset = intent_logits[:, start_index:end_index]

            intent_loss += self.objective(subset, labels[:, slot])
            predicted_intent.append(subset.max(1)[1])
            start_index = end_index

        predicted_intent = torch.stack(predicted_intent, dim=1)
        records['acc'] += (predicted_intent == labels).prod(1).view(-1).cpu().float().tolist()
        records['intent_loss'].append(intent_loss.item())

        def idx2slots(indices: torch.Tensor):
            action_idx, object_idx, location_idx = indices.cpu().tolist()
            return (
                self.Sy_intent["action"][action_idx],
                self.Sy_intent["object"][object_idx],
                self.Sy_intent["location"][location_idx],
            )

        records["filename"] += filenames
        records["predict"] += list(map(idx2slots, predicted_intent))
        records["truth"] += list(map(idx2slots, labels))

        return intent_loss

    # interface
    def log_records(self, mode, records, logger, global_step, **kwargs):
        save_names = []
        for key in ["acc", "intent_loss"]:
            values = records[key]
            average = torch.FloatTensor(values).mean().item()
            logger.add_scalar(
                f'fluent_commands/{mode}-{key}',
                average,
                global_step=global_step
            )
            with open(Path(self.expdir) / "log.log", 'a') as f:
                if key == 'acc':
                    print(f"{mode} {key}: {average}")
                    f.write(f'{mode} at step {global_step}: {average}\n')
                    if mode == 'dev' and average > self.best_score:
                        self.best_score = torch.ones(1) * average
                        f.write(f'New best on {mode} at step {global_step}: {average}\n')
                        save_names.append(f'{mode}-best.ckpt')

        with open(Path(self.expdir) / f"{mode}_predict.csv", "w") as file:
            lines = [f"{f},{a},{o},{l}\n" for f, (a, o, l) in zip(records["filename"], records["predict"])]
            file.writelines(lines)

        with open(Path(self.expdir) / f"{mode}_truth.csv", "w") as file:
            lines = [f"{f},{a},{o},{l}\n" for f, (a, o, l) in zip(records["filename"], records["truth"])]
            file.writelines(lines)

        return save_names



import logging
from torch.nn.utils.rnn import pad_sequence

from ..interfaces import UpstreamBase

logger = logging.getLogger(__name__)


class UpstreamExpert(UpstreamBase):
    def __init__(self, ckpt, model_config, **kwargs):
        super().__init__(**kwargs)

        try:
            from pase.models.frontend import wf_builder

        except ModuleNotFoundError:
            logger.error(
                "Please check https://github.com/s3prl/s3prl/blob/master/s3prl/upstream/pase/README.md "
                "for how to install the dependencies of PASE+."
            )
            raise

        def build_pase(ckpt, model_config):
            pase = wf_builder(model_config)
            pase.load_pretrained(ckpt, load_last=True, verbose=False)
            return pase

        self.model = build_pase(ckpt, model_config)

    def get_downsample_rates(self, key: str) -> int:
        return 160

    def forward(self, wavs):
        wavs = pad_sequence(wavs, batch_first=True)
        wavs = wavs.unsqueeze(1)

        features = self.model(wavs)  # (batch_size, feature_dim, extracted_seqlen)
        features = features.transpose(
            1, 2
        ).contiguous()  # (batch_size, extracted_seqlen, feature_dim)

        # This forward function only does the model forward
        # The return dict is then handled by UpstreamBase's hooks



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ utility/get_best_score.py ]
#   Synopsis     [ scripts to find the best score from training logs ]
#   Author       [ Andy T. Liu (Andi611) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


"""
Usage:
    `python utility/get_best_score.py result/downstream/example/log.log dev test +`

This will find the `test` score ranked by the best `dev` score.
+: higher is better
-: lower is better
"""


###############
# IMPORTATION #
###############
import sys


########
# MAIN #
########
def main():

    log_file = str(sys.argv[1])
    
    if len(sys.argv) == 4:
        rank_by = str(sys.argv[2])
        target = str(sys.argv[3])
        large_or_small = str(sys.argv[4])
    else:
        rank_by = 'dev'
        target = 'test'
        large_or_small = '+'

    best_record = [-99999, 0, None]
    if large_or_small == '-': best_record[0] *= -1

    with open(log_file) as f:
        lines = f.readlines()
        
        for line in lines:
            line = line.strip('\n').split('/')[-1].split('|')
            if len(line) < 3: continue

            prefix = str(line[0].split(':')[-1])
            step = int(line[1].split(':')[-1])
            score = float(line[2].split(':')[-1])

            if rank_by in prefix:
                if compare(score, best_record[0], large_or_small):
                    best_record[0] = score # the score to rank by
                    best_record[1] = step
            
            elif step == best_record[1] and target in prefix:
                best_record[2] = score # the score you want to find
    
    print(f'The best {rank_by} score {best_record[0]} at step {best_record[1]}, accoupanied by this {target} score {best_record[2]}')


def compare(a, b, large_or_small):
    if large_or_small == '+':
        return a > b
    elif large_or_small == '-':
        return a < b
    else:
        raise ValueError(large_or_small)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ expert.py ]
#   Synopsis     [ the any-to-one voice conversion downstream wrapper ]
#   Author       [ Wen-Chin Huang (https://github.com/unilight) ]
#   Copyright    [ Copyright(c), Toda Lab, Nagoya University, Japan ]
"""*********************************************************************************************"""


import os
import numpy as np
from scipy.io.wavfile import write
from sklearn.preprocessing import StandardScaler
from tqdm import tqdm
import yaml

import torch
import torch.nn as nn
from torch.utils.data import DataLoader, DistributedSampler
from torch.distributed import is_initialized
from torch.nn.utils.rnn import pad_sequence

from .model import Model
from .dataset import VCC2020Dataset
from .utils import make_non_pad_mask
from .utils import read_hdf5, write_hdf5
from .utils import logmelspc_to_linearspc, griffin_lim

FS = 16000

class Loss(nn.Module):
    """
    L1 loss module supporting (1) loss calculation in the normalized target feature space
                              (2) masked loss calculation
    """
    def __init__(self, stats):
        super(Loss, self).__init__()
        self.objective = torch.nn.L1Loss(reduction="mean")
        self.register_buffer("target_mean", torch.from_numpy(stats.mean_).float())
        self.register_buffer("target_scale", torch.from_numpy(stats.scale_).float())

    def normalize(self, x):
        return (x - self.target_mean) / self.target_scale

    def forward(self, x, y, x_lens, y_lens, device):
        # match the input feature length to acoustic feature length to calculate the loss
        if x.shape[1] > y.shape[1]:
            x = x[:, :y.shape[1]]
            masks = make_non_pad_mask(y_lens).unsqueeze(-1).to(device)
        if x.shape[1] <= y.shape[1]:
            y = y[:, :x.shape[1]]
            masks = make_non_pad_mask(x_lens).unsqueeze(-1).to(device)
        
        # calculate masked loss
        x_normalized = self.normalize(x)
        y_normalized = self.normalize(y.to(device))
        x_masked = x_normalized.masked_select(masks)
        y_masked = y_normalized.masked_select(masks)
        loss = self.objective(x_masked, y_masked)
        return loss

class DownstreamExpert(nn.Module):
    def __init__(self, upstream_dim, upstream_rate, downstream_expert, expdir, **kwargs):
        super(DownstreamExpert, self).__init__()
        
        # basic settings
        self.expdir = expdir
        self.upstream_dim = upstream_dim
        self.trgspk = downstream_expert['trgspk']
        self.datarc = downstream_expert['datarc']
        self.modelrc = downstream_expert['modelrc']
        self.acoustic_feature_dim = self.datarc["fbank_config"]["n_mels"]
        self.fs = self.datarc["fbank_config"]["fs"]
        self.resample_ratio = self.fs / self.datarc["fbank_config"]["n_shift"] * upstream_rate / FS
        print('[Downstream] - resample_ratio: ' + str(self.resample_ratio))

        # load datasets
        self.train_dataset = None

        # load statistics file if exists, and calculate if not found
        scaler = StandardScaler()
        stats_root = self.datarc["stats_root"]
        if not os.path.exists(stats_root):
            os.makedirs(stats_root)
        stats_path = os.path.join(stats_root, self.trgspk+".h5")
        if os.path.exists(stats_path):
            print("[Stats] - reading stats from " + str(stats_path))
            scaler.mean_ = read_hdf5(stats_path, "mean")
            scaler.scale_ = read_hdf5(stats_path, "scale")
        else:
            print("[Stats] - " + str(stats_path) + " does not exist. Calculating statistics...")
            self.train_dataset = VCC2020Dataset('train', self.trgspk, **self.datarc)
            for _, _, lmspc, _ in self.train_dataset:
                scaler.partial_fit(lmspc)
            write_hdf5(stats_path, "mean", scaler.mean_.astype(np.float32))
            write_hdf5(stats_path, "scale", scaler.scale_.astype(np.float32))
            print("[Stats] - writing stats to " + str(stats_path))
        self.stats = scaler

        # define model and loss
        self.model = Model(
            input_dim = self.upstream_dim,
            output_dim = self.acoustic_feature_dim,
            resample_ratio = self.resample_ratio,
            stats = self.stats,
            **self.modelrc
        )
        self.objective = Loss(self.stats)


    # Interface
    def get_dataloader(self, split):
        if split == 'train':
            if self.train_dataset is None:
                self.train_dataset = VCC2020Dataset('train', self.trgspk, **self.datarc)
            return self._get_train_dataloader(self.train_dataset)            
        elif split == 'dev':
            self.dev_dataset = VCC2020Dataset('dev', self.trgspk, **self.datarc)
            return self._get_eval_dataloader(self.dev_dataset)
        elif split == 'test':
            self.test_dataset = VCC2020Dataset('test', self.trgspk, **self.datarc)
            return self._get_eval_dataloader(self.test_dataset)
        elif split == 'custom_test':
            from .dataset import CustomDataset
            return self._get_eval_dataloader(CustomDataset(self.datarc["eval_list_file"]))


    def _get_train_dataloader(self, dataset):
        sampler = DistributedSampler(dataset) if is_initialized() else None
        return DataLoader(
            dataset, batch_size=self.datarc['train_batch_size'],
            shuffle=(sampler is None),
            sampler=sampler,
            num_workers=self.datarc['num_workers'],
            collate_fn=dataset.collate_fn
        )


    def _get_eval_dataloader(self, dataset):
        return DataLoader(
            dataset, batch_size=self.datarc['eval_batch_size'],
            shuffle=False, num_workers=self.datarc['num_workers'],
            collate_fn=dataset.collate_fn
        )


    # Interface
    def forward(self,
                split,
                input_features,
                wavs,
                acoustic_features,
                acoustic_features_padded,
                acoustic_feature_lengths,
                wav_paths,
                records,
                **kwargs):

        device = input_features[0].device

        # padding
        input_feature_lengths = torch.IntTensor([feature.shape[0] for feature in input_features])
        input_features = pad_sequence(input_features, batch_first=True).to(device=device)
        
        # forward model
        if split == "custom_test":
            predicted_features, predicted_feature_lengths = self.model(input_features, input_feature_lengths)
            records["predicted_features"] += predicted_features.cpu().numpy().tolist()
            records["feature_lengths"] += predicted_feature_lengths.cpu().numpy().tolist()
            records["wav_paths"] += wav_paths
            return 0.0 # return dummy value
        else:
            if split in ["dev", "test"]:
                predicted_features, predicted_feature_lengths = self.model(input_features, input_feature_lengths)
                # save the unnormalized features for dev and test sets
                records["predicted_features"] += predicted_features.cpu().numpy().tolist()
                records["feature_lengths"] += predicted_feature_lengths.cpu().numpy().tolist()
                records["wav_paths"] += wav_paths
                records["wavs"] += wavs
            else:
                predicted_features, predicted_feature_lengths = self.model(input_features, input_feature_lengths, acoustic_features_padded.to(device))

            # loss calculation (masking and normalization are done inside)
            loss = self.objective(predicted_features,
                                acoustic_features_padded,
                                predicted_feature_lengths,
                                acoustic_feature_lengths,
                                device)
            records['loss'].append(loss.item())

            return loss

    # interface
    def log_records(self, split, records, logger, global_step, batch_ids, total_batch_num, **kwargs):
        if split == "custom_test":
            hdf5_save_dir = os.path.join(self.expdir, split, "hdf5")
            wav_save_dir = os.path.join(self.expdir, split, "wav")
            os.makedirs(hdf5_save_dir, exist_ok=True)
            os.makedirs(wav_save_dir, exist_ok=True)

            for i, wav_path in enumerate(tqdm(records["wav_paths"], dynamic_ncols=True, desc="Saving files")):
                length = int(records["feature_lengths"][i])
                fbank = np.array(records["predicted_features"][i])[:length]

                # save generated features into hdf5 files
                hdf5_save_path = os.path.join(hdf5_save_dir, (self.trgspk + "_" + os.path.basename(wav_path)).replace(".wav", ".h5"))
                write_hdf5(hdf5_save_path, "feats", fbank)

                # mel fbank -> linear spectrogram
                spc = logmelspc_to_linearspc(
                    fbank,
                    fs=self.datarc["fbank_config"]["fs"],
                    n_mels=self.datarc["fbank_config"]["n_mels"],
                    n_fft=self.datarc["fbank_config"]["n_fft"],
                    fmin=self.datarc["fbank_config"]["fmin"],
                    fmax=self.datarc["fbank_config"]["fmax"],
                )
                # apply griffin lim algorithm
                y = griffin_lim(
                    spc,
                    n_fft=self.datarc["fbank_config"]["n_fft"],
                    n_shift=self.datarc["fbank_config"]["n_shift"],
                    win_length=self.datarc["fbank_config"]["win_length"],
                    window=self.datarc["fbank_config"]["window"],
                    n_iters=self.datarc["fbank_config"]["gl_iters"],
                )
                # save generated waveform
                wav_save_path = os.path.join(wav_save_dir, self.trgspk + "_" + os.path.basename(wav_path))
                write(
                    wav_save_path,
                    self.datarc["fbank_config"]["fs"],
                    (y * np.iinfo(np.int16).max).astype(np.int16),
                )
        else:    
            loss = torch.FloatTensor(records['loss']).mean().item()

            if split in ["dev", "test"]:

                hdf5_save_dir = os.path.join(self.expdir, str(global_step), split, "hdf5")
                wav_save_dir = os.path.join(self.expdir, str(global_step), split, "wav")
                os.makedirs(hdf5_save_dir, exist_ok=True)
                os.makedirs(wav_save_dir, exist_ok=True)

                for i, wav_path in enumerate(tqdm(records["wav_paths"])):
                    length = int(records["feature_lengths"][i])
                    fbank = np.array(records["predicted_features"][i])[:length]

                    # save generated features into hdf5 files
                    hdf5_save_path = os.path.join(hdf5_save_dir, "_".join(wav_path.split("/")[-2:]).replace(".wav", ".h5"))
                    write_hdf5(hdf5_save_path, "feats", fbank)

                    # mel fbank -> linear spectrogram
                    spc = logmelspc_to_linearspc(
                        fbank,
                        fs=self.datarc["fbank_config"]["fs"],
                        n_mels=self.datarc["fbank_config"]["n_mels"],
                        n_fft=self.datarc["fbank_config"]["n_fft"],
                        fmin=self.datarc["fbank_config"]["fmin"],
                        fmax=self.datarc["fbank_config"]["fmax"],
                    )
                    # apply griffin lim algorithm
                    y = griffin_lim(
                        spc,
                        n_fft=self.datarc["fbank_config"]["n_fft"],
                        n_shift=self.datarc["fbank_config"]["n_shift"],
                        win_length=self.datarc["fbank_config"]["win_length"],
                        window=self.datarc["fbank_config"]["window"],
                        n_iters=self.datarc["fbank_config"]["gl_iters"],
                    )
                    # save generated waveform
                    wav_save_path = os.path.join(wav_save_dir, "_".join(wav_path.split("/")[-2:]))
                    write(
                        wav_save_path,
                        self.datarc["fbank_config"]["fs"],
                        (y * np.iinfo(np.int16).max).astype(np.int16),
                    )

            print(f'{split} loss: {loss:.6f}')

            save_names = []
            logger.add_scalar(
                f'example/{split}-{loss}',
                loss,
                global_step=global_step
            )
            return save_names
